/* 
 * ConfigurationManager.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.core;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;
import ants.util.LineListTokenizer;
import ants.util.LineListException;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.Flow;
import edu.utah.janos.nodeos.FlowRun;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.io.Log;

import edu.utah.janosvm.kit.comm.commspace.CommSpaceHandle;

import java.io.InputStreamReader;
import java.io.StringReader;
import java.io.FileNotFoundException;
import java.io.IOException;

import java.util.Enumeration;
import java.util.Vector;

/**
 * A configuration manager that creates a node interface,
 * channel and app configuration.  Gets the configuration
 * from a control file.
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class ConfigurationManager
    extends FlowRun
{
    /**
     * Default port for channel connections.
     */
    public static final String DEFAULT_PORT = "4242";

    /**
     * Parse a single command.
     * 
     * @param node the node to attach created entities to
     * @param command array of command and its argument words
     * @exception AddressFormatException if an address parameter is bad
     * @exception ArrayIndexOutOfBoundsException if insufficent paramters are given for some commands.
     * @exception AttachException if attaching Channel/Extension/App fails
     * @exception ClassNotFoundException if any dynamically loaded class isn't found.
     * @exception IllegalAccessException if newInstance() fails
     * @exception InstantiationException if newInstance() fails
     * @exception InvalidKeyArgsException if any optional args are improperly formatted
     */
    private final void parseCommandLine(Vector appVector, String[] command)
	throws 
	AddressFormatException,
	ArrayIndexOutOfBoundsException,
	AttachException,
	ClassNotFoundException,
	IllegalAccessException,
	InstantiationException,
	InvalidKeyArgsException
    {
	Trace.trace(Trace.ANTS_STARTUP, "Parsing command " +command[0]);
	
	/*
	 * Format of all command lines:
	 * 	<command> <address> [<args> ...]
	 * The first two elements are required.  The args are dependent
	 * upon the command.
	 */
	if (command[0].equals("node")) {
	    // FORMAT: node <node addr> [<optional arg> ...]
	    
	    // Set address before calling setArgs so that routeTable will exist.
	    PrimordialNode.setAddress(Address.fromString(command[1]));

	    /*
	     * Primordial Node handles all the optional arguments for
	     * the node command.
	     */
	    PrimordialNode.setArgs(new KeyArgs(command, 2));
	    
	} else if (command[0].equals("channel")) {
	    /*
	     * channel <nodeaddr> <recvaddr> [-proxyaddr <proxyaddr>]
	     * 
	     * <nodeaddr> is the same as before, the node's active address
	     * <recvaddr> is the address where packets for this channel will
	     * 	      be received by the node.  If no proxy is specified,
	     * 	      this is also where remote nodes should send packets to.
	     * 	      The format for this field is:
	     * 		      'stdin' or <host>[:<port>]
	     * [-proxyaddr <proxyaddr>] is the address to which remote nodes should
	     * 	      send packets destined to this node.  The format of <proxyaddr>
	     * 	      is <host>[:<port>].
	     */

	    if (command.length < 3)
		throw new ArrayIndexOutOfBoundsException("insufficent "
							 + "arguments for "
							 + "channel command");
	    
	    String recvAddr = command[2];
	    String proxyAddr = null;

	    int i = 3;
	    while (i < command.length)
	    {
		// XXX assume all optional arguments come in pairs
		if (i == command.length)
		    throw new InvalidKeyArgsException("Option " +command[i]+
						      " missing required argument.");

		if (command[i].equals("-proxyaddr")) 
		{
		    i++;
		    proxyAddr = command[i];
		}
		else if (command[i].equals("-log"))
		{
		    // XXX legacy option, IGNORED
		    i++;
		}
		else
		{
		    throw new InvalidKeyArgsException("Unknown option " +command[i]);
		}

		i++;
	    }
	    
	    // Convert recvAddr into a protocol and address string for
	    // the NodeOS.

	    String protocolSpec;
	    String addrSpec;
	    
	    if (recvAddr.equalsIgnoreCase("stdin"))
	    {
		protocolSpec = "stdio:anep";
		addrSpec = "";
	    }
	    else
	    {
		String host = "localhost";
		String portStr = DEFAULT_PORT;
		
		int sep = recvAddr.indexOf(':');
		if (sep != 0)
		{
		    host = recvAddr.substring(0,sep);
		    portStr = recvAddr.substring(sep+1, recvAddr.length());
		}
		else
		{
		    host = recvAddr;
		}

		protocolSpec = "if/ipv4/udp";
		addrSpec = "*:" +portStr+ ":" +host;
	    }

	    // Convert proxyAddr into a protocol and address spec for
	    // the NodeOS.

	    String proxyProtoSpec = null;
	    String proxyAddrSpec = null;
	    
	    if (proxyAddr != null)
	    {
		String host = "localhost";
		String portStr = DEFAULT_PORT;

		int sep = proxyAddr.indexOf(':');
		if (sep != 0)
		{
		    host = proxyAddr.substring(0,sep);
		    portStr = proxyAddr.substring(sep+1, proxyAddr.length());
		}
		else
		{
		    host = proxyAddr;
		}

		proxyProtoSpec = "if/ipv4/udp";
		proxyAddrSpec = "*:" +portStr+ ":" +host;
	    }

	    // Store the proxy and physical addrs in the PrimordialNode.

	    Trace.trace(Trace.ANTS_STARTUP, "Proxy addr: (" +proxyProtoSpec+ " ; " +proxyAddrSpec+ ")");
	    Trace.trace(Trace.ANTS_STARTUP, "Phys addr: (" +protocolSpec+ " ; " +addrSpec+ ")");

	    PrimordialNode.setProxyAddress(proxyProtoSpec, proxyAddrSpec);
	    PrimordialNode.setPhysicalAddress(protocolSpec, addrSpec);
	} else if (command[0].equals("nchannel")) {
	    /*
	     * nchannel <nodeaddr> <protocol> <recvaddr> [-proxyaddr <protocol> <proxyaddr>]
	     * 
	     * <nodeaddr> is the same as before, the node's active address
	     * <protocol> is the protocol spec to use (usally if/ipv4/udp).
	     * <recvaddr> is the address where packets for this channel will
	     * 	      be received by the node.  If no proxy is specified,
	     * 	      this is also where remote nodes should send packets to.
	     * 	      The format for this field is:
	     *         <src address>:<dest port>:<dest address>:<src port>
	     * [-proxyaddr <protocol> <proxyaddr>] is the address to which
	     *        remote nodes should send packets destined to this node.
	     *        The format of <proxyaddr> is
	     *        <src address>:<dest port>:<dest address>:<src port>
	     */

	    if (command.length < 4)
		throw new ArrayIndexOutOfBoundsException("insufficent "
							 + "arguments for "
							 + "channel command");

	    String protocolSpec = command[2];
	    String addrSpec = command[3];
	    String proxyProtoSpec = null;
	    String proxyAddrSpec = null;

	    int i = 4;
	    while (i < command.length)
	    {
		if (command[i].equals("-proxyaddr")) 
		{
		    if (i >= command.length-1)
			throw new InvalidKeyArgsException("Option " +command[i]+
							  " missing required argument.");
		    i++;
		    proxyProtoSpec = command[i];
		    i++;
		    proxyAddrSpec = command[i];
		}
		else if (command[i].equals("-log"))
		{
		    if (i == command.length)
			throw new InvalidKeyArgsException("Option " +command[i]+
							  " missing required argument.");
		    // XXX legacy option, IGNORED
		    i++;
		}
		else
		{
		    throw new InvalidKeyArgsException("Unknown option " +command[i]);
		}

		i++;
	    }
	    
	    // Store the proxy and physical addrs in the PrimordialNode.

	    Trace.trace(Trace.ANTS_STARTUP, "Proxy addr: (" +proxyProtoSpec+ " ; " +proxyAddrSpec+ ")");
	    Trace.trace(Trace.ANTS_STARTUP, "Phys addr: (" +protocolSpec+ " ; " +addrSpec+ ")");

	    PrimordialNode.setProxyAddress(proxyProtoSpec, proxyAddrSpec);
	    PrimordialNode.setPhysicalAddress(protocolSpec, addrSpec);
	} else if (command[0].equals("application")) {
	    // FORMAT: application <node addr> <app class> [<optional args> ...]
	    
	    if (command.length < 3)
		throw new ArrayIndexOutOfBoundsException("insufficient arguments for application");
	    
	    /*
	     * Store the name of the app, it will be forced into a
	     * new Flow and started up in Node.start()
	     */
	    appVector.addElement(new AppConfig(command[2], command, 3));
	    
	} else if (command[0].equals("extension")) {
	    PrimordialNode.message(Log.ERR, "extension command ignored by ANTS 2.0.");  // XXX
	    // IGNORE.  Not used in ANTS v2.0
	    // FORMAT: extension <node addr> <class name> [<optional args> ... ]
	    
	} else if (command[0].equals("manager")) {
	    PrimordialNode.message(Log.ERR, "manager command ignored by ANTS 2.0.");  // XXX
	    // IGNORE.  Not used in ANTS 2.0
	    
	} else if (command[0].equals("connect")) {
	    // IGNORE.  Its a legit command that is handled by makeroutes...
	    
	} else {
	    PrimordialNode.message(Log.ERR, "Ignored bad command " +command[0]+ "."); // XXX
	}
    }
    
    
    /**
     */
    private final void parseCommands(Vector apps, String[] params)
    {
	// two required args
	String filename = params[0];
	String nodename = params[1];
	
	// Remainder go in a KeyArgs structure
	try
	{
	    KeyArgs k = new KeyArgs(params, 2);
	    int lpc;

	    for( lpc = 0; lpc < k.length(); lpc++ )
	    {
		if( k.key(lpc).equals("--debug") )
		{
		    Trace.setFlags(k.arg(lpc));
		    k.strike(lpc);
		}
	    }
	}
	catch (InvalidKeyArgsException ikae)
	{
	    Node.abort(ikae, "ConfigurationManager args are bogus.");
	}
	
	/*
	 * Process the configuration file, creating entities
	 * as needed for each command.
	 */
	
	LineListTokenizer lt = null;
	try
	{
	    lt = new LineListTokenizer(filename);
	}
	catch (FileNotFoundException fnfe)
	{
	    Node.abort(fnfe, "ConfigurationManager cannot load file " +filename );
	}
	
	Trace.trace(Trace.ANTS_STARTUP, "Parsing config file " +filename);

	String[] args = null;
	do
	{
	    try
	    {
		args = lt.list();
	    }
	    catch (LineListException lle)
	    {
		Node.abort(lle, "malformed line in " +filename+ ": " +lle.getMessage());
	    }
	    
	    if (args == null)
		continue;
	    
	    if (args.length < 2)
	    {
				// XXX do real i/o
		PrimordialNode.message(Log.ERR,
				       "Bad command/line "
				       + args[0]
				       + ".");
		continue;
	    }
	    
	    // Ignore commands that aren't for this node.
	    if (! args[1].equals(nodename))
		continue;
	    
	    try
	    {
		parseCommandLine(apps, args);
	    }
	    catch (Exception e)
	    {
		Node.abort(e, "When parsing command " +args[0]+ ".");
	    }
	}
	while (args != null);
    }
    
    /**
     * OLD way to start an ANTS node.  Does not apply in ANTS 2.0.
     *
     * @see #run(Flow,String[],CommSpaceHandle)
     */
    public static void main(String[] args)
	throws Exception
    {
	PrimordialNode.message(Log.ERR,
			       "ConfigurationManager must be started as a Janos Flow.");
	System.exit(11);
    }
    
    /**
     * Start a new configuration manager.
     * Prints a usage message of insufficient arguments are given.
     * <p>
     * This is the entrypoint from the NodeOS at startup.
     *
     * @param flow the root flow for this EE to run in
     * @param args the command line arguments to the ConfigurationManager
     * @param commSpaceHandle for bootstrapping inter-process communication
     * @see #parseCommands(Vector,String[])
     * @see PrimordialNode#start(Vector,Vector)
     * @see FlowRun
     */
    protected void run(Flow flow,
		       String[] args,
		       CommSpaceHandle commSpaceHandle)
    {
	Assert.require(flow == Flow.currentFlow());
	
	flow.setName("Primoridal ANTS Flow");
	
	if ((args == null)
	    || (args.length < 2))
	{
	    PrimordialNode.message(Log.WARNING,
				   "usage: " 
				   +ConfigurationManager.class.getName()+
				   "<file> <node> [--debug <flags>]");
	    flow.destroy();
	}
	
	Vector apps = new Vector();
	Vector exts = new Vector();
	parseCommands(apps, args);
	
	Trace.trace(Trace.ANTS_STARTUP, "Config file command parsing complete.");

	try
	{
	    PrimordialNode.start(apps, exts);
	}
	catch (Exception e)
	{
	    PrimordialNode.message(Log.ERR, "Exception in PrimordialNode startup: " +e.getMessage());
	    PrimordialNode.dumpStackTrace(Log.ERR, e);
	    PrimordialNode.shutdown();
	}
    }
}
