/*
 * Console.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import edu.utah.janos.io.Log;

import ants.util.LineListTokenizer;
import ants.util.LineListException;

import java.net.ServerSocket;
import java.net.Socket;

import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.StringReader;
import java.io.PrintWriter;

/**
 * The console creates a connection point for remote telnet clients
 * to connect to to control an ANTS node.  After the console is create
 * a thread must be dedicated to it via the <code>run()</code> method
 * to handle incoming connections.  A <code>ConsoleWorker</code> is forked off for
 * each incoming connection.
 * <p>
 * XXX add a clean shutdown method
 *
 * @see ConsoleWorker
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
class Console
    implements java.lang.Runnable // XXX
{
    private static final int MAX_SOCKET_BACKLOG = 2;
    
    /**
     * The socket this console listens on.
     */
    private final ServerSocket socket;
    
    private final Node node;
    
    /**
     * Create a Console listener on the given port.
     * Blows up if the port is already in use, throwing
     * an IOException.
     * <p>
     * Connections will not be handled until the <code>accept()</code>
     * method is executed.  You'd best supply a new thread
     * to handle them, too.
     *
     * @see #accept()
     * @param port the port to use; zero for any free port.
     * @exception IOException throws whatever errors ServerSocket might toss your way.
     */
    Console(Node node, int port)
	throws java.io.IOException // XXX 
    {
	this.node = node;
	
	// May throw IOException:
	this.socket = new ServerSocket(port, MAX_SOCKET_BACKLOG);
    }
    
    /**
     * Return the port this console server listens on
     */
    int getPort()
    {
	return socket.getLocalPort();
    }
    
    public String toString()
    {
	return "Console[" +socket.toString()+ "]";
    }
    
    /**
     * Set the connection handler function in motion
     * <p>
     * This function never returns.
     */
    public void run()
    {
	accept();
    }
    
    /**
     * Accept and handle connections on this Console.
     * <p>
     * This call never returns.
     */
    void accept()
    {
	ConsoleWorker worker;
	Socket client;
	
	while (true)
	{
	    try
	    {
		client = this.socket.accept();
		worker = new ConsoleWorker(this.node, client);
		this.node.systemThreadStart(worker, "ConsoleWorker");
	    }
	    catch(Throwable th)
	    {
		node.log(Log.WARNING,
			 "Exception on console: "
			 + th.getMessage());
		node.logStackTrace(Log.WARNING, th);
	    }
	}
    }
}
