/*
 * NeighborOutChannel.java
 *
 * Copyright (c) 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core;

import java.io.IOException;

import edu.utah.janos.nodeos.Flow;
import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.OutChannel;
import edu.utah.janos.nodeos.ProtocolSpecFormatException;
import edu.utah.janos.nodeos.PacketAddressFormatException;

/**
 * The NeighborOutChannel is used as an indirection record a binding between an
 * out channel and a neighbor address.  It can also be used to detect when the
 * address has changed and a new out channel needs to be created.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a> */
class NeighborOutChannel
{
    /**
     * The address of the neighbor.
     */
    private Address addr;

    /**
     * An OutChannel that should point to the neighbor.
     */
    private OutChannel oc;

    /**
     * Construct a NeighborOutChannel that refers to the given neighbor.
     *
     * @param addr The address of the neighbor.
     */
    NeighborOutChannel(Address addr)
	throws IOException
    {
	this.setNeighbor(addr);
    }

    /**
     * Reset the address of the neighbor and recreate the out channel to it.
     *
     * @param addr The address of the neighbor.
     * @exception Throws an IOException if it failed to create the OutChannel.
     */
    void setNeighbor(Address addr)
	throws IOException
    {
	if( this.oc != null )
	{
	    this.oc.destroy();
	    this.oc = null;
	}

	this.addr = addr;
	this.oc = this.createOutChannel();
    }

    /**
     * Create an OutChannel pointing to the neighbor.
     *
     * @exception Throws an IOException if it failed to create the OutChannel.
     */
    private OutChannel createOutChannel()
	throws IOException
    {
	OutChannel retval;

	try
	{
	    retval = OutChannel.create(Flow.currentFlow(),
				       this.addr.getProtocolSpec(),
				       this.addr.getAddressSpec());
	}
	catch(PacketAddressFormatException e)
	{
	    throw new InternalError();
	}
	catch(ProtocolSpecFormatException e)
	{
	    throw new InternalError();
	}
	return retval;
    }

    /**
     * @return The OutChannel that points at the neighbor.
     */
    OutChannel getOutChannel()
    {
	return this.oc;
    }

    /**
     * Destroy the OutChannel that this object refers too.
     */
    void destroy()
    {
	if( this.oc != null )
	{
	    this.oc.destroy();
	    this.oc = null;
	}
    }

    /**
     * Validate that the currently known address of a neighbor matches the
     * one bound to the out channel.
     *
     * @param addr The currently known address of the neighbor.
     * @return True if the given address matches the one bound to the out
     * @return channel.
     */
    boolean valid(Address addr)
    {
	return (this.addr == addr);
    }

    public String toString()
    {
	return "NeighborOutChannel[addr="
	    + this.addr
	    + "; oc="
	    + this.oc
	    + "]";
    }
}
