/* 
 * NodeCache.Java
 *  
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 * 
 */
package ants.core;

import edu.utah.janos.nodeos.Flow;
import edu.utah.janos.util.LinkedList;

/**
 * Node cache of managed soft-state.
 * <p>
 * Combines hashtable and doubly linked list to implement fixed size table
 * with LRU replacement.  all public entry points are synchronized.
 * a background gc thread cleans entries that have expired.
 * <p>
 * XXX the background GC thread is probably unnecessary since this is
 * only per-flow state.  At the least, the background thread should
 * be configurable by the flow (e.g., its collection parameters).
 * <p>
 * XXX should this class be final?  Should more methods be marked
 * final? (e.g., get)
 * <p>
 * This differs from the original version in that it does not have a maximum
 * time limit or a maximum number of cached items, both of which are unneeded
 * because of per-flow resource limits.
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class NodeCache
    implements Runnable
{
    /**
     * Entries in the hashtable and lru list.
     * Can be nested under NodeCache in 1.1
     */
    private final class NCElement
	extends LinkedList.Node
    {
	/**
	 * Link to next hash table elments
	 */
	NCElement next;
	
	/**
	 * Key for this node
	 */
	Object key;
	
	/**
	 * Hash code for this key, faster to compare than the actual key
	 */
	int hash;
	
	/**
	 * Stored value for the key
	 */
	Object value;
	
	/**
	 * Lifetime of the key/value pair
	 */
	long expiration;
    }
    
    /**
     * Elements of this type in the node cache will receive calls when
     * one of a set of events occurs.
     * <p>
     * XXX Not sure if this is useful or not, and it can probably be done
     * better
     */
    public abstract static class EventHandler
    {
	/**
	 * This is called when the element is manually removed from
	 * from the cache.
	 */
	public void removed()
	{
	}
	
	/**
	 * This is called when this element is up for collection,
	 * the return value specifies the new expiration time.  Zero
	 * is returned if the element should really be collected
	 */
	public int collected()
	{
	    return 0;
	}
    }
    
    /**
     * Reference to whichever node owns this cache
     */
    private final Node owner;
    
    /**
     * Hash table for quick access to the NCElements
     */
    private NCElement[] hash;
    
    /**
     * Doubly linked list for tracking the least recently used elements in
     * case we want to kill them before their times have expired
     */
    private final LinkedList lru = new LinkedList();
    
    /**
     * Tracks the number of elements;
     */
    private int elementCount;
    
    /**
     * The minimum expiration time of the current set of elments.  This
     * value is relative and is used as a timeout for
     * java.lang.Object.wait()
     */
    private long minTime;
    
    /**
     * Tell the gc to continue collecting
     */
    private boolean keepCollecting = true;
    
    /**
     * Get the bucket for a specific hash code
     *
     * @param hcode A hash code, most likely produced by
     * java.lang.Object.hashCode()
     */
    private final int hashBucket( int hcode )
    {
	int retval;
	
	retval = (hcode & 0x7FFFFFF) % this.hash.length;
	return( retval );
    }
    
    /**
     * Find the NCElement with the given key.  The NCElement will also
     * be moved to the front of the LRU, and hash table lists.
     *
     * @param key The key to match on
     * @return The NCElement matching the given key
     */
    private synchronized NCElement find(Object key)
    {
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	NCElement nce, pred = null;
	NCElement retval = null;
	
	for( nce = this.hash[bucket];
	     (nce != null) && (retval == null);
	     pred = nce, nce = nce.next )
	{
	    /*
	     * Compare the hash code first so we don't have to
	     * compare the whole key.  (Questionable whether this
	     * is a good thing or not)
	     */
	    if( (nce.hash == hcode) && nce.key.equals(key) )
	    {
		retval = nce;
				/*
				 * Push the hash element to the front for
				 * faster access
				 */
		if( pred != null )
		{
		    pred.next = nce.next;
		    nce.next=this.hash[bucket];
		    this.hash[bucket] = nce;
		}
				/*
				 * Push the element to the front of the list
				 */
		if( this.lru.getHead() != nce )
		{
		    nce.remove();
		    this.lru.addHead(nce);
		}
	    }
	}
	return retval;
    }
    
    /**
     * Return the value paired with a certain key
     *
     * @param key The key to search for
     * @return The value associated with the key
     */
    public final synchronized Object get(Object key)
    {
	Object retval = null;
	NCElement nce;
	
	nce = this.find(key);
	if( nce != null )
	    retval = nce.value;
	return retval;
    }
    
    /**
     * Add a key/value pair to the cache with a given expiration time.
     * If the key is already present in the cache then it will be replaced
     * with the parameters and the old value returned.
     *
     * @param key The key for identifying the value
     * @param value The value to be tracked
     * @param time The expiration time for this association
     * @return Old value from a prior association, if it existed
     */
    public final synchronized Object put(Object key,
					 Object value,
					 int time)
    {
	Object retval = null;
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	NCElement nce;
	
	nce = this.find(key);
	
	if( nce == null )
	{
	    /* This is a new key, so we create an element for it */
	    nce = new NCElement();
	    nce.hash = hcode;
	    nce.key = key;
	    nce.next = this.hash[bucket];
	    this.hash[bucket] = nce;
	    this.lru.addHead( nce );
	    this.elementCount++;
	}
	else
	    retval = nce.value;
	nce.value = value;
	nce.expiration = this.owner.time() + time * 1000;
	/*
	 * If the expiration for this association happens before the
	 * current minimum then we need to wake up the gc thread and
	 * have it wait for the new time.
	 */
	if( (time * 1000) < this.minTime )
	{
	    synchronized(this.lru)
	    {
		this.lru.notify();
	    }
	}
	return retval;
    }
    
    /**
     *
     */
    public synchronized boolean refresh(Object key, int time)
    {
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	boolean retval = false;
	NCElement nce;
	
	nce = this.find(key);
	if( nce != null )
	{
	    retval = true;
	    nce.expiration = this.owner.time() + time * 1000;
	    if( (time * 1000) < this.minTime )
	    {
		synchronized(this.lru)
		{
		    this.lru.notify();
		}
	    }
	}
	return retval;
    }
    
    /**
     * Remove an association from the cache.
     *
     * @param key Key for the association to remove
     * @return The value from the association
     */
    public synchronized Object remove(Object key)
    {
	Object retval = null;
	
	int hcode = key.hashCode();
	int bucket = this.hashBucket( hcode );
	
	NCElement e = this.hash[bucket];
	NCElement prev = null;
	
	while( (e != null) && (retval == null) )
	{
	    if( (e.hash == hcode) && e.key.equals(key) )
	    {
		if (prev != null)
		    prev.next = e.next;
		else
		    this.hash[bucket] = e.next;
		e.next = null;
		retval = e.value;
		e.remove();
		this.elementCount--;
		if( e.value instanceof EventHandler )
		{
		    ((EventHandler)e.value).removed();
		}
				/*
				 * Its not a big deal if this object has the
				 * minimum time used for the gc thread.  It
				 * just won't collect anything, and anyways
				 * it will find the new minTime from the scan.
				 */
	    }
	    else
	    {
		prev = e;
		e = e.next;
	    }
	}
	return retval;
    }
    
    /**
     * Overloaded get method for integer keys.
     */
    public final synchronized Object get(int i)
    {
	return this.get(new Integer(i));
    }
    
    /**
     * Overloaded put method for integer keys.
     */
    public final synchronized Object put(int i, Object value, int time)
    {
	return this.put(new Integer(i), value, time);
    }
    
    /**
     * Overloaded get method for long keys.
     */
    public final synchronized Object get(long l)
    {
	return this.get(new Long(l));
    }
    
    /**
     * Overloaded put method for long keys.
     */
    public final synchronized Object put(long l, Object value, int time)
    {
	return this.put(new Long(l), value, time);
    }
    
    /**
     * Scans the list of associations and garbage collects based on
     * expiration dates.
     */
    private synchronized void clean()
    {
	long now = this.owner.time();
	LinkedList.Node curr, pred;
	NCElement nce;
	
	/*
	 * We have to determine a new min time since we might be
	 * collecting the current one
	 */
	this.minTime = Long.MAX_VALUE;
	/*
	 * Scan backwards through the list so we can trash excess stuff
	 * based on the LRU.
	 */
	curr = this.lru.getTail();
	while( curr.getPred() != null )
	{
	    pred = curr.getPred();
	    nce = (NCElement)curr;
	    if(nce.expiration <= now)
	    {
		if( nce.value instanceof EventHandler )
		{
		    int newtime =
			((EventHandler)nce.value).
			collected();
		    
		    nce.expiration =
			this.owner.time()
			+ newtime
			* 1000;
		    if( newtime == 0 )
			this.remove(nce.key);
		    this.minTime = Math.min(this.minTime,
					    nce.expiration - now);
		}
		else
		    this.remove(nce.key);
	    }
	    else
	    {
		this.minTime = Math.min(this.minTime,
					nce.expiration - now);
	    }
	    curr = pred;
	}
    }
    
    /**
     * Run forever, collecting expired objects.
     */
    public void run()
    {
	while( this.keepCollecting )
	{
	    /* Wait for a while... */
	    try
	    {
		synchronized(this.lru)
		{
				/*
				 * If there aren't any elements we can
				 * just wait for forever, and let put()
				 * wake us up.  Otherwise, we wait
				 * until the next association expires.
				 */
		    if( this.elementCount == 0 )
			this.lru.wait();
		    else
			this.lru.wait(this.minTime);
		}
	    }
	    catch( InterruptedException e )
	    {
		;
	    }
	    /* Garbage collect, and then wait some more... */
	    this.clean();
	}
    }
    
    /**
     * Start the collection thread
     *
     * @param max The maximum number of elements allowed
     */
    void collect()
    {
	/*
	 * XXX should this be systemThreadStart?  
	 * If it were, then users could create lots of NodeCache()
	 * to create lots of threads...
	 */
	this.owner.threadStart(this, "NodeCache Garbage Collector");
    }
    
    void stopCollecting()
    {
	this.keepCollecting = false;
	synchronized(this.lru)
	{
	    this.lru.notify();
	}
    }
    
    /**
     * Construct a new NodeCache.  Fires up the thread associated with
     * this object.
     * <p>
     * XXX this probably shouldn't be public.
     *
     * @param owner The Node that owns this NodeCache instance
     * @param size The size of the hash table
     * @param duration Maximum allowable expiration time for an association
     */
    public NodeCache(Node owner, int size)
    {
	this.owner = owner;
	this.hash = new NCElement[size];
	this.minTime = Long.MAX_VALUE;
    }
    
    public String toString()
    {
	return "NodeCache[elementCount="
	    + this.elementCount
    + "; minTime="
    + this.minTime
    + "]";
    }
}
