/*
 * RouteTable.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.core;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.AddressFormatException;
import edu.utah.janos.nodeos.NoSuchNodeAddress;

import edu.utah.janos.route.RouteEntry;
import edu.utah.janos.route.WriteableRouteTable;

import ants.core.Assert;
import ants.core.Trace;

import java.util.Enumeration;

// XXX:
import java.io.FileInputStream;
import java.io.StreamTokenizer;
import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.io.IOException;


/**
 * The basic ANTS route table.
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class RouteTable
    extends WriteableRouteTable
{
    /**
     * Initialize the route table and the neighbor table
     * from the given routeFile.  A null routeFile implies
     * that there are no pre-defined routes.  nodeAddr may
     * be zero, implying the current node has no address (yet).
     * If nodeAddr is zero, routeFile must be null.
     *
     * @param nodeAddr the address of this node 0 if this node
     * does not have an address yet.
     * @param node the node for the primordial flow
     * @param routeFile the name of the file to load and read
     * route information from.  May be null.
     * @param routeTable an instance of route table to fill in
     * with data from the file.
     * @param neighborTable an instance of a neighbor table to
     * fill in with data from the file.
     */
    public static void initializeTables(int nodeAddr,
					String routeFile,
					RouteTable routeTable,
					NeighborTable neighborTable)
	throws IOException, AddressFormatException, NoSuchNodeAddress
    {
	Assert.require(routeTable);
	Assert.require(neighborTable);
	Assert.require((nodeAddr != 0) || (routeFile == null));

	// no routes to load is okay
	if (routeFile == null)
	    return;

	StreamTokenizer st = new StreamTokenizer(
	    new BufferedReader(
		new InputStreamReader(
		    new FileInputStream(routeFile))));
	
	// Initialize the tokens for the stream reader
	st.resetSyntax();
	st.whitespaceChars(0, ' ');
	st.wordChars('0', '9');
	st.wordChars('.', '.');
	st.wordChars(':', ':');
	st.wordChars('-', '-');
	st.wordChars('A', 'Z');
	st.wordChars('a', 'z');
	st.commentChar('#');
	
	int src = 0;
	int dst = 0;
	int nxt = 0;
	String addr = null;
	
	Trace.trace(Trace.ANTS_ROUTE_READ_TABLE,
		    "Reading route table '" +routeFile+
		    "' for node " +Address.toString(nodeAddr));
	
	while (st.nextToken() == StreamTokenizer.TT_WORD)
	{
	    if (src == 0)
	    {
		Assert.assert(dst == 0);
		Assert.assert(nxt == 0);
		Assert.assert(addr == null);
		src = Address.fromString(st.sval);
	    }
	    else if (dst == 0)
	    {
		Assert.assert(nxt == 0);
		Assert.assert(addr == null);
		dst = Address.fromString(st.sval);
	    } 
	    else if (nxt == 0)
	    {
		Assert.assert(addr == null);
		nxt = Address.fromString(st.sval);
	    }
	    else
	    {
		Assert.assert(addr == null);
		addr = st.sval;
		
		Trace.trace(Trace.ANTS_ROUTE_READ_TABLE,
			    "Adding route " +Address.toString(src)+
			    " -> " +Address.toString(dst)+
			    " via " +Address.toString(nxt)+ "(" +addr+ ")");
		
		// If this line is for a route from this node
		if (src == nodeAddr) 
		{
		    RouteEntry entry = new RouteEntry(dst,nxt,addr);

		    // XXX addRoute inserts a clone, be careful.
		    routeTable.addRoute(entry);
 		    
		    // check if this is a gateway
		    if (dst == nxt)
		    {
			neighborTable.addNeighbor(dst, entry.getPhysicalNext());
		    }
		}
		
		// Clear values for next line of route file
		src = 0;
		dst = 0;
		nxt = 0;
		addr = null;
	    }
	}
	
	if (src != 0) 
	    throw new IOException("premature end of input"); // xxx wrong exeception
	else if (st.ttype != StreamTokenizer.TT_EOF)
	    throw new IOException("can't handle " + st.toString());// xxx wrong exeception
    }
}
