/* 
 * BasicPolicy.java
 *
 * Copyright (c) 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.core.security;

import java.util.Hashtable;
import java.util.Vector;

/**
 * A hashtable-based implementation of a security policy that
 * maps principals to permissions. 
 *
 * @author Andrew Whitaker
 */
public class BasicPolicy
    extends Policy
    implements java.io.Serializable
{
    private Hashtable acl;
    
    /**
     * Creates a blank policy 
     */
    public BasicPolicy()
    {
	acl = new Hashtable();
    }
    
    /**
     * Perform an access control check.  This method uses the implies
     * method of the Permission class, so "read write" implies the
     * "read" permission.
     *
     * @see Permission
     */
    public boolean hasPermission(Principal who, Permission what)
    {
	Vector v;
	
	if (who == null)
	    throw new NullPointerException("Null Principal passed to BasicPolicy.hasPermission()");
	if (what == null)
	    throw new NullPointerException("Null Permission passed to BasicPolicy.hasPermission()");
	
	v = (Vector) acl.get(who);
	if (v == null)
	    return false;
	else
	{
	    for (int i=0;i<v.size();i++)
	    {
		if ( ((Permission) v.elementAt(i)).implies(what))
		    return true;
	    }
	}
	
	return false;
    }
    
    /**
     * Add a permission after first checking for exact duplicates 
     */
    public void addPermission(Principal who, Permission what)
    {
	Vector v = (Vector) acl.get(who);
	if (v == null)
	{
	    Vector v2 = new Vector();
	    v2.addElement(what);
	    acl.put(who,v2);
	}
	else
	{
	    if (! (v.contains(what)))
		v.addElement(what);
	}
    }
    
    /**
     * Remove a permission from a given principal.  This method uses the
     * equals method of the Permission class, so only exact matches are
     * removed.  That is, "read write" does not remove "read"
     */
    public void removePermission(Principal who, Permission what)
    {
	Vector v = (Vector) acl.get(who);
	if (v == null)
	    return;
	else v.removeElement(what);
    }
    
    /**
     * Remove all permissions for the given principal 
     */
    public void removePrincipal(Principal who)
    { 
	acl.remove(who);
    }
}



