/*
 * DynamicRouteTable.Java
 * 
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2000, 2001 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.route;

import ants.core.Application;
import ants.core.Capsule;
import ants.core.Node;
import ants.core.Xdr;
import ants.core.NeighborTable;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

import edu.utah.janos.io.Log;

import edu.utah.janos.nodeos.Address;
import edu.utah.janos.nodeos.NoSuchNodeAddress;
import edu.utah.janos.nodeos.AddressFormatException;

import java.util.Vector; // XXX
import java.util.Hashtable; // XXX
import java.util.Enumeration; // XXX

import ants.core.Trace;
import ants.core.Assert;


/**
 * The "app" that maintains a node's dynamic routing tables.
 * <p>
 * This is really just an app so it can justify sending capsules off
 * without priming them off some other capsule.  (All this code used
 * to be part of DynamicRouteTable.)
 * <p>
 * Understands the default '-port' argument that all apps grok.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @author <a href="mailto:djw@cs.washington.edu">David Wetherall</a>
 */
public class DynamicRouteApp
    extends Application
    implements ants.core.NeighborListener
{
    public static final int APP_PORT = 13;
    
    private static final String[] defaults = {"-port", Integer.toString(APP_PORT)};
    
    /**
     * Maximum number of route updates to stuff into a
     * single RouteUpdateCapsule.  (Each update costs 5 bytes).
     */
    private final static int MAXROUTESPERUPDATE = 64;
    
    /**
     * The table this app is keeping up to date.
     */
    private DynamicRouteTable table = null;
    
    /**
     * A table of neighboring nodes
     */
    private NeighborTable neighbors;
    
    /**
     * The node this table is maintained on.
     */
    private Node node;
    
    /**
     * Used for suppressing triggered updates.  Triggered
     * updates are scheduled when the routing table is
     * changed.  Once a triggered update is scheduled,
     * futher updates are suppressed until the trigger
     * update is executed.
     */
    private boolean suppressTriggerEvents;
    
    /**
     * The queue of timers and events that cause updates.
     */
    private final DynamicRouteEvents events;
    
    /**
     * App constructors can't do anything useful.  Ideally,
     * I would like to set this.node and this.table, but I
     * can't because the node hasn't been plugged into this
     * app yet, and I need to node to get the table...
     */
    public DynamicRouteApp()
    {
	this.events = new DynamicRouteEvents();
	this.suppressTriggerEvents = false;
    }
    
    /**
     * Parse command line arguments.
     * <p>
     * See the class description for a list of the arguments and
     * their meaning.
     */
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	super.setArgs(k);
    }
    
    /**
     * Initialize the event queue.
     */
    private void initEvents()
    {
	long now = node.time();
	
	for (Enumeration e = table.getRoutes(); e.hasMoreElements();) 
	{
	    DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();
	    events.setExpireTimer(r, now);

	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Initial route: " +r);
	}
    }
    
    /**
     * Do initialization of the App.
     * <p>
     * XXX one-time function.
     */
    private void initApp()
    {
	this.node = thisNode();
	Assert.assert(this.node);
	
 	this.table = new DynamicRouteTable(node.getRouteTable());
	
	this.neighbors = node.getNeighborTable();
	this.neighbors.addNeighborListener(this);
	
	Assert.assert(this.table);
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "DynamicRouteApp starting.");
	
	/*
	 * Install the update protocol (register capsules, etc).
	 */
	try
	{
	    DynamicRouteProtocol rp = new DynamicRouteProtocol();
	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Installing " +rp);
	    node.register(rp, node.getPrincipal());
	}
	catch (Exception e)
	{
	    // XXX wrong exception
	    throw new InternalError("DynamicRouteProtocol setup failure:: "
				    +e.getClass().getName() 
				    +": "
				    +e.getMessage());
	}
	
	/*
	 * Initialize events from the gateways setup in the routing table
	 */
	initEvents(); // XXX one-time function
    }
    
    private void printRouteTable(String prefix)
    {
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, prefix+ "Current Route Table: ");
	boolean donesomething = false;
	
	for (Enumeration e = table.getRoutes();e.hasMoreElements();) 
	{
	    DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();
	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "   " +r);
	    donesomething = true;
	}

	if (!donesomething)
	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "   <no routes>");
    }

    /**
     * The active route table logic.
     * <p>
     * Harmless if invoked on a static table.
     */
    public void start()
    {
	/*
	 * Setup local state that couldn't be set in the constructor.
	 */
	initApp();
	
	// Synchronize on this table.  Is that sufficient??  
	// Drops lock in table.wait, below...
	synchronized (table)
	{
	    // Install next periodic wakeup
	    events.setPeriodicTimer(node.time());
	    
	    while (true) 
	    {
		final DynamicRouteEvent event;
		
		Assert.assert(!events.isEmpty());
		
		// find the next scheduled event
		event = events.findFirstTimer();
		
		// printRouteTable("");

		// if it's time, then process it
		long now = node.time();
		long delay = event.when - now;
		if (delay <= 0)
		{
		    Trace.trace(Trace.ANTS_ROUTE_TIMER, "Handling event."
				+ " now=" +now
				+ " event=" +event);

		    // sanity check for leftovers in the queue
		    Assert.assert(event.what != DynamicRouteEvent.EMPTY); // XXX
		    
		    // Each event may have more than one duty.  Do all that apply.
		    
		    if ((event.what & DynamicRouteEvent.GC) != 0) {
			gcEntries(event);
		    }
		    
		    if ((event.what & DynamicRouteEvent.EXPIRE) != 0) {
			expireEntries(event);
		    }
    
		    // Note that route messages are sent if the event is
		    // a periodic or trigger timer.  Don't waste work.
		    if ((event.what & DynamicRouteEvent.PERIODIC) != 0) {
			sendRouteMessages(DynamicRouteEvent.PERIODIC);
			events.setPeriodicTimer(event.when);
			
		    } else if ((event.what & DynamicRouteEvent.TRIGGER) != 0) {
			sendRouteMessages(DynamicRouteEvent.TRIGGER);
			clearChangeFlags();
		    }
		    
		    // clear trigger even if we had a PERIODIC+TRIGGER timer.
		    if ((event.what & DynamicRouteEvent.TRIGGER) != 0)
			suppressTriggerEvents(false);
		    
		    // Toss the event, we've handled it.
		    events.clearTimer(event);
		}
		else
		{
		    /*
		     * First event is far in the future.  Block until
		     * that day arrives.
		     */
		    try
		    {
			Trace.trace(Trace.ANTS_ROUTE_TIMER, "RouteTable: Waiting for " + delay + "ms ...");
			
			// *wait* (not sleep) so that we release the lock on the table.
			table.wait(delay);
		    }
		    catch (InterruptedException e)
		    {
			// Spurious wakeups are harmless
			// XXX statcount this
			;
		    }
		}
	    }
	    
	    // NOT REACHED
	}
    }
    
    
    // -- BEGIN LOCAL ROUTE TABLE PROCESSING ---
    
    /**
     * Walks all routes and moves expired routes into the gcPool.
     */
    private void expireEntries(DynamicRouteEvent e)
    {
	boolean change;
	do // XXX why loop until nothing changes?  Always true on second iteration?
	{
	    change = false;
	    for (Enumeration f = table.getRoutes(); f.hasMoreElements();)
	    {
		DynamicRouteEntry r = (DynamicRouteEntry)f.nextElement();
		if (r.getExpiration() == e.when)
		{
		    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "DynamicRouteApp: Route to "+
				Address.toString(r.getLogicalDest()) +" (gw="+
				Address.toString(r.getLogicalNext()) + ") expired.");
		    
		    // Move route to GC Pool.
		    r.setCost(table.INFINITY);
		    table.gcPool().put(new Integer(r.getLogicalDest()), r);
		    table.removeRoute(r);
		    r.change = true;
		    events.setGCTimer(r, r.getExpiration());
		    
		    scheduleTrigger(r.getExpiration());
		    change = true;
		    
		    break;
		}
	    }
	} while (change);
    }
    
    /**
     * Delete all routes in the gcPool that have expired.  Such routes
     * are lost for all time.
     */
    private void gcEntries(DynamicRouteEvent e)
    {
	boolean change;
	
	do // XXX why is this loop here?
	{
	    change = false;
	    for (Enumeration f = table.gcPool().elements(); f.hasMoreElements();)
	    {
		DynamicRouteEntry r = (DynamicRouteEntry)f.nextElement();
		if (r.getExpiration() == e.when) {
		    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "RouteTable: GCing route to " + Address.toString(r.getLogicalDest()));
		    table.gcPool().remove(new Integer(r.getLogicalDest()));
		    change = true;
		    break;
		}
	    }
	} while (change);
    }
    
    /**
     * Mark all routes as 'unchanged'.
     */
    private void clearChangeFlags()
    {
	for (Enumeration f = table.getRoutes(); f.hasMoreElements();)
	{
	    DynamicRouteEntry d = (DynamicRouteEntry)f.nextElement();
	    d.change = false;
	}
	
	for (Enumeration f = table.gcPool().elements(); f.hasMoreElements();)
	{
	    DynamicRouteEntry d = (DynamicRouteEntry)f.nextElement();
	    d.change = false;
	}
    }
    
    /**
     * Send a route update to the given node.
     * <p>
     * XXX this is completely out of synch with the new world order.
     */
    private void sendUpdate(int oneHopNeighbor, Vector r)
    {
	boolean sentIt;
	DynamicRouteCapsule cap = 
	    new DynamicRouteCapsule(oneHopNeighbor,
				    node.getVisibleProtocolSpec(),
				    node.getVisibleAddressSpec(),
				    r);
	
	// XXX may cause recursion into the
	// RouteTable.. shouldn't be bad, though.
	sentIt = send(cap);
	
	if (!sentIt)
	{
	    // XXX probably means that oneHopNeighbor has disappeared.
	    // node.log(Log.WARNING, "Failed to send DynamicRouteCapsule " +cap+ ".  Perhaps destination left network?");
	}
	
    }
    
    /**
     * Send a message to a host about a route.  The given
     * host should be a gateway host for this node.
     */
    private void sendRouteWorker(int oneHopNeighbor, int cause)
    {
	final Vector routeChunk = new Vector( MAXROUTESPERUPDATE ); // XXX hit that limit often?
	boolean update = false;
	boolean gcswitch = false;
	int routes = 0;
	int poison = 0;
	boolean gwchange = false;
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "sendRouteWorker(neighbor="
		    +Address.toString(oneHopNeighbor)
		    +", cause="+cause+")");

	for (Enumeration f = table.getRoutes(); f.hasMoreElements();)
	{
	    DynamicRouteEntry currRoute = (DynamicRouteEntry)f.nextElement(); // xxx check cast
	    
	    // search the gc pool after the main table... oi.
	    if (!gcswitch & !f.hasMoreElements())
	    {
		f = table.gcPool().elements();
		gcswitch = true;
	    }
	    
	    // skip the route to that gateway.  Record a route
	    // change to the gateway
	    if (currRoute.getLogicalDest() == oneHopNeighbor)
	    {
		if (currRoute.change)
		    gwchange = true;
		continue;
	    }
	    
	    // Don't spew unchanged route on a trigger event
	    if ((cause == DynamicRouteEvent.TRIGGER)
		&& (!currRoute.change))
		continue;
	    
	    // Route is changed or this is a periodic update.
	    
	    // Poison all routes through the current gateway.  It knows
	    // better than we do the cost of the path... 
	    int cost = currRoute.getCost();
	    if (currRoute.getLogicalNext() == oneHopNeighbor)
	    {
		cost = table.INFINITY;
		poison++;
	    }
	    
	    DynamicRouteEntry u = new DynamicRouteEntry(currRoute);
	    u.setCost(cost);
	    routeChunk.addElement(u);
	    routes++;
	    
	    // If we've hit high-water mark, send what we've got and continue afresh.
	    if (routes == MAXROUTESPERUPDATE) {
		
		Trace.trace(Trace.ANTS_ROUTE_UPDATE, "RouteTable: Sending " + routes + " routes ("+ poison + " poison) to " + Address.toString(oneHopNeighbor));
		
		sendUpdate(oneHopNeighbor, routeChunk);
		routeChunk.removeAllElements();
		routes = 0;
		poison = 0;
		update = true;
	    }
	}
	
	// Send final chunk of route updates.  For periodic
	// timers, make sure we sent at least one update (even if its empty).
	// Or, if the gateway changed on a trigger, send an empty update.
	if ((routes > 0)
	    || (!update && (cause == DynamicRouteEvent.PERIODIC))
	    || (gwchange && (cause == DynamicRouteEvent.TRIGGER)))
	{
	    if (routes>0)
	    {
		Trace.trace(Trace.ANTS_ROUTE_UPDATE, "RouteTable: Sending " + routes + " routes ("+ poison + " poison) to " + Address.toString(oneHopNeighbor));
		sendUpdate(oneHopNeighbor, routeChunk);
	    } else {
		Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Warning: No active routes");
	    }
	    routeChunk.removeAllElements();
	    routes = 0;
	    poison = 0;
	}
    }
    
    /**
     * Send a message to each gateway and to host which
     * has a cost of "1" to reach.  For each gateway send
     * a long list of route entries.
     * <p>
     * XXX this is O(N * N) when N is the *number of routes*!
     *
     * @param cause the reason this was invoked (Trigger or periodic).
     */
    private void sendRouteMessages(int cause)
    {
	Assert.assert((cause == DynamicRouteEvent.TRIGGER)
		      || (cause == DynamicRouteEvent.PERIODIC));
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "sendRouteMessages(cause="+cause+")");

	// generate messages for each directly connected gateway
	for (Enumeration e = neighbors.getNeighbors();
	     e.hasMoreElements(); /* */)
	{
	    //DynamicRouteEntry r = (DynamicRouteEntry)e.nextElement();
	    Integer i = (Integer) e.nextElement();
	    sendRouteWorker(i.intValue(), cause);
	}
    }
    
    /**
     * Receive and process RouteUpdateCapsule.
     * RouteUpdateCapsules contain zero or more route update
     * records.  Each "record" consists of a destination
     * and a "distance" to that destination.
     *
     * @param unknownCap the RouteUpdateCapsule.
     */
    public void receive(Capsule unknownCap) {
	DynamicRouteCapsule cap;
	
	if (!(unknownCap instanceof DynamicRouteCapsule))
	    return;
	
	cap = (DynamicRouteCapsule)unknownCap;
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Got an update capsule " +cap);
	
	/*
	 * Note that update capsules should *never* have to forward
	 * themselves because they're only sent to 1-hop-away
	 * gateways.
	 */
	if (node.getAddress() != cap.getDst())
	{
	    node.log(Log.WARNING,
		     "DynamicRouteApp received a DynamicRouteCapsule destined for"
		     + Address.toString(node.getAddress())
		     + " instead of "
		     + Address.toString(cap.getDst())
		     + "(me).  Dropping capsule.");
	    return;
	}
	
	boolean changed = false;
	long now = node.time();
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "RouteUpdateCapsule: Arrived from " + Address.toString(cap.getSrc()) + " ("+ cap.nroutes + " routes)");
	
	/*
	 * Synchronize on the table.  We also synch on the table
	 * up in the start() method.
	 */
	synchronized(table)
	{ 
	    DynamicRouteEvent n1 = events.findFirstTimer();
	    String gwAddr = repairPhysAddr(cap.protocolSpec + ";" + cap.addressSpec);
	    
	    if (newRouteInfo(cap.getSrc(), 0, cap.getSrc(), gwAddr, now))
		changed = true;
	    
	    if (cap.nroutes > 0)
	    {
		Xdr xdr = new Xdr(cap.routes, 0);
		
		for (int i = 0; i < cap.nroutes; i++) {
		    int dest = xdr.INT();
		    int dist = xdr.BYTE();
		    
		    if (newRouteInfo(dest, dist, cap.getSrc(), gwAddr, now))
			changed = true;
		}
	    }
	    
	    if (changed)
	    {
		// schedule a trigger update
		scheduleTrigger(now);
	    }
	    
	    // wake the scheduler if we've altered the next event
	    DynamicRouteEvent n2 = events.findFirstTimer();
	    if (n2.when < n1.when)
		table.notify();
	}
	return;
    }
    
    /**
     * The physical address passed into newRouteInfo is in a different
     * format than the route table expects it to be, so it need to be
     * fixed.
     *
     * @param gwPhyAddr The physical address in format proto;port:port:host
     * @return The physical address in format host:port
     */
    private String repairPhysAddr(String gwPhysAddr)
    {
	String physaddr;
	// Parse the gwPhysAddr into what the route expects
	physaddr=gwPhysAddr.substring(gwPhysAddr.indexOf(":")+1);
	// Now the two items need to be reversed :P
	return physaddr.substring(physaddr.indexOf(":")+1)+":"+
	    physaddr.substring(0,physaddr.indexOf(":"));
    }
    
    /**
     * Provide new information for a single route.  If the info
     * is useful, update the route table.  It is the caller's
     * responsibility to schedule a trigger update if there are
     * changes to the route table.
     * <p>
     * Invoked by exclusivly by <code>DynamicRouteCapsule.evalutate</code>.
     *
     * @param dest the active address of the destination to update a route to
     * @param dist the distance to that node from the gateway
     * @param gw the active address of the gateway (the "next hop" to dest)
     * @param gwPhysAddr the physical address of the gateway
     * @param now what time is it my dear?
     * @return true if a change was made; false if not.
     * @see DynamicRouteCapsule#evaluate(Node)
     */
    boolean newRouteInfo(int dest, int dist,
			 int gw, String gwPhysAddr,
			 long now)
    {
	DynamicRouteEntry oldRoute;
	int cost = dist + 1;
	
	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Updating route: to " +Address.toString(dest)
		    + " via " + Address.toString(gw) + "(" +gwPhysAddr + ")"
		    + " cost " + cost);
	
	// Snap all costs at infinity.
	if (cost > table.INFINITY)
	    cost = table.INFINITY;
	
	oldRoute = (DynamicRouteEntry)table.getRoute(new Integer(dest));

	if (oldRoute == null)
	{
	    DynamicRouteEntry newRoute;
	    
	    // Bail on the expensive ones...
	    if (cost == table.INFINITY)
		return false;
	    
	    try
	    {
		newRoute = new DynamicRouteEntry(dest, gw, gwPhysAddr);
		newRoute.setCost(cost);
		newRoute.change = true;
		
	    }
	    catch (NoSuchNodeAddress nsne)
	    {
		node.log(Log.WARNING, "Route update with bad address: " +nsne.getMessage());
		return false;
	    }
	    catch (AddressFormatException afe)
	    {
		node.log(Log.WARNING, "Route update with bad address: " +afe.getMessage());
		return false;
	    }
	    
	    // might have to resuscitate from gc pool
	    DynamicRouteEntry gc = (DynamicRouteEntry)table.gcPool().get(new Integer(dest));
	    if (gc != null)
	    {
		events.clearGCTimer(gc.getExpiration());
		table.gcPool().remove(new Integer(dest));
	    }
	    
	    events.setExpireTimer(newRoute, now);
	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "  Added clean new route " +newRoute);

	    // XXX addRoute inserts a clone.  Be careful.
	    table.addRoute(newRoute);
	    
	    // Made changes
	    return true;
	}
	// route exists, see if anything changed as per input processing spec
	else if ((((oldRoute.getLogicalNext() == gw) && (cost != oldRoute.getCost())))
		 || (cost < oldRoute.getCost()))
	{
	    DynamicRouteEntry newRoute;
	    
	    try
	    {
		newRoute = new DynamicRouteEntry(dest, gw, gwPhysAddr);
		newRoute.setCost(cost);
		newRoute.change = true;
	    }
	    catch (NoSuchNodeAddress nsne)
	    {
		node.log(Log.WARNING, "Route update with bad address: " +nsne.getMessage());
		return false;
	    }
	    catch (AddressFormatException afe)
	    {
		node.log(Log.WARNING, "Route update with bad address: " +afe.getMessage());
		return false;
	    }
	    
	    
	    // Found in live pool, cannot be in gc pool
	    Assert.assert(table.gcPool().get(new Integer(dest)) == null);
	    
	    /*
	     * XXX new world order: cannot update a RouteEntry.  
	     * (Specifically, if the gw has changed).  Thus,
	     * have to remove and re-insert.
	     */
	    
	    events.clearExpireTimer(oldRoute.getExpiration());
	    table.removeRoute(oldRoute);
	    
	    if (cost == table.INFINITY)
	    {
		// infinite links go right into the gc pool
		table.gcPool().put(new Integer(dest), newRoute);
		newRoute.change = true;
		events.setGCTimer(newRoute, newRoute.getExpiration());
	    }
	    else
	    {
		events.setExpireTimer(newRoute, now);
		// XXX addRoute inserts a clone.  Be careful.
		table.addRoute(newRoute);
	    }
	    
	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "  Replaced old route with " +newRoute);
	    
	    return true;
	}      
	
	// finally, just reset the timers if confirming update
	else if ((oldRoute.getLogicalNext() == gw)
		 && (cost == oldRoute.getCost()))
	{
	    // Found in live pool, cannot be in gc pool
	    Assert.assert(table.gcPool().get(new Integer(dest)) == null);
	    
	    events.clearExpireTimer(oldRoute.getExpiration());
	    events.setExpireTimer(oldRoute, now);
	}
	
	// if we reach here, there were no changes
	return false;
    }
    
    // -- END ROUTEUPDATECAPSULE INTERFACES --
    
    
    /**
     * Suppress scheduling of further trigger events.  Or
     * don't.
     */
    void suppressTriggerEvents(boolean doit)
    {
	suppressTriggerEvents = doit;
    }
    
    /**
     * Schedule a trigger event for some time after 'when'.
     * If triggers are surpressed (because one has already been
     * scheduled but not executed) the request is dropped on the floor.
     * <p>
     * Invoked in here and by RouteUpdateCapsule.
     */
    void scheduleTrigger(long when)
    {
	if (!suppressTriggerEvents)
	{
	    events.setTriggerTimer(when);
	    suppressTriggerEvents = true;
	}
    }
    
    // --- The NeighborListener Interface

    /**
     * A neighbor was added to the neighbor table
     */
    public void neighborAdded(int address, Address link)
    {
	DynamicRouteEntry dre;
	long now = node.time();
	long oldFirstEvent;

	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Add neighbor "
		    +Address.toString(address)+
		    " link=" +link);

	dre = new DynamicRouteEntry(address, address, link);
	dre.setCost(1);
	dre.change = true;

	synchronized(table)
	{
	    events.setExpireTimer(dre, now);

	    // XXX addRoute inserts a clone.  Be careful.
	    table.addRoute(dre);

	    Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Added new neighbor route " +dre);

	    // wake scheduler if we've altered the first event.
	    // XX this is bougs.  Should be done by scheduleTrigger...
	    oldFirstEvent = events.findFirstTimer().when;
	    scheduleTrigger(dre.getExpiration());
	    if (oldFirstEvent > events.findFirstTimer().when)
		table.notify();
	}
    }
    
    /**
     * A neighbor was removed from the neighbor table
     */
    public void neighborRemoved(int address, Address link)
    {
	DynamicRouteEntry dre;

	Assert.require(this.table);
	Assert.require(link);

	dre = (DynamicRouteEntry)table.getRoute(new Integer(address));

	Trace.trace(Trace.ANTS_ROUTE_UPDATE, "Remove neighbor "
		    +Address.toString(address)+
		    "link=" +link+ "; (route entry = " +dre+ ")");

	// XXX what does it mean if dre is null?  Removing an already
	// gone route?

	if ((dre != null)
	    && dre.getPhysicalNext().equals(link))
	{
	    synchronized(table)
	    {
		table.removeRoute(dre);
		table.notify();
	    }
	}
    }
}
