/* 
 * DynamicRouteEvent.Java
 *
 * Copyright (c) 1999, 2001 The University of Utah and the Flux Group.
 * Copyright (c) 2000 The University of Washington
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.route;

import ants.core.Assert;

/**
 * A timer event affecting the DynamicRouteTable entries.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 * @author David Wetherall
 */
class DynamicRouteEvent 
{
    // different types of route events:
    
    /**
     * An EMPTY event.  DynamicRouteEvents are never
     * initialized with this value.  They just
     * decay to it after they expire.
     */
    final static int EMPTY    = 0;
    
    /**
     * The PERODIC event.  These trigger periodically (duh)
     * and send a full update to all neighbor gateways.
     */
    final static int PERIODIC = 1<<0;
    
    /**
     * The TRIGGER event.  Trigger events are scheduled for
     * a short time in the future to propagate recent
     * changes.  
     */
    final static int TRIGGER  = 1<<1;
    
    /**
     * The GC event is triggered on a route when it is
     * time for the route to be retired.  Routes are
     * first moved into a gc pool where they can be
     * ressurected.  Otherwise they will EXPIRE.
     */
    final static int GC       = 1<<2;
    
    /**
     * The EXPIRE event.  This event is triggered on a route
     * after its been placed in the GC pool and if it
     * doesn't get resurrected.  EXPIRED routes are deleted.
     */
    final static int EXPIRE   = 1<<3;
    
    /**
     * Timestamp of when this event is scheduled to
     * be invoked.
     */
    final long when;
    
    /**
     * A set of flags indicating what sort of events this
     * object implies (Periodic, trigger, expire, or GC).
     * <p>
     * This is mutable.  Events may be added if they occur
     * at the same time as a different event.
     */
    int what;
    
    /**
     * Let me know if you figure out what this does.
     */
    int expireCount;
    
    /**
     * Let me know if you figure out what this does.
     */
    int gcCount;
    
    DynamicRouteEvent(long when, int what) 
    {
	Assert.assert(what != EMPTY);
	Assert.assert(when > 0);
	
	this.when = when;
	this.what = what;
	this.expireCount = 0;
	this.gcCount = 0;
    }

    public String toString()
    {
	StringBuffer whatStr = new StringBuffer(128);
	boolean append = false;

	if ((what&PERIODIC) != 0)
	{
	    if (append)
		whatStr.append(",");
	    whatStr.append("PERIODIC");
	    append=true;
	}
	if ((what&TRIGGER) != 0)
	{
	    if (append)
		whatStr.append(",");
	    whatStr.append("TRIGGER");
	    append=true;
	}
	if ((what&GC) != 0)
	{
	    if (append)
		whatStr.append(",");
	    whatStr.append("GC");
	    append=true;
	}
	if ((what&EXPIRE) != 0)
	{
	    if (append)
		whatStr.append(",");
	    whatStr.append("EXPIRE");
	    append=true;
	}

	if (whatStr == null)
	    whatStr.append("EMPTY");

	return "DynamicRouteEvent[when=" +when+
	    "; what=" +whatStr
	    + (((this.what&EXPIRE) != 0)
	       ? "; expireCt=" +expireCount : "")
	    + (((this.what&GC) != 0)
	       ? "; gcCount=" +gcCount : "")
	    + "]";
    }
}

