/*
 * ClassPathReader.Java
 *
 * Copyright (c) 1998,1999,2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 */
package ants.util;

import java.util.Vector;
import java.util.Hashtable;
import java.util.zip.ZipFile;
import java.util.zip.ZipEntry;
import java.io.File;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.IOException;

/**
 * This class implements the ByteCodeReader interface
 * based on a "CLASSPATH" which can be given as a parameter.
 * <p>
 * XXX throws 'Exception' in a bunch of places.   Should have its own
 * exception class.
 * <p>
 * XXX no longer used (everyone uses the Janos Java NodeOS APIs now).
 *
 * @deprecated Replaced by the Janos Java NodeOS class loading facilities.
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class ClassPathReader
    implements ByteCodeReader
{
    /**
     * Components of this path
     */
    private Vector pathcomps;
    
    /**
     * Hashtable of already read classes
     */
    private Hashtable	classcache;
    
    
    /**
     * Use the system property 'java.class.path' as the default
     * classpath to search.
     */
    public ClassPathReader() 
    {
	this(System.getProperty("java.class.path"));
    }
    
    /**
     * Build a class path from the specified path string
     *
     * @param pathstr the CLASSPATH to use for future class searches.
     */
    public ClassPathReader(String pathstr) 
    {
	classcache = new Hashtable();
	pathcomps = new Vector(5,2);
	
	String s = pathstr;
	int i;
	while ((i = s.indexOf(File.pathSeparatorChar)) != -1)
	{
	    if (i == 0)
		pathcomps.addElement(".");
	    else
		pathcomps.addElement(s.substring(0, i));
	    s = s.substring(i+1);
	}
	pathcomps.addElement(s);
    }
    
    /**
     * Resolve a class name of the form foo.bar.baz
     * Searches the classpath given to the constructor.
     *
     * @param classname fully qualified class name
     * @return byte[] array of bytes containing .class file
     */
    public byte[] readByteCode(String classname) throws Exception
    {
	// check in cache first
	byte [] p = (byte [])classcache.get(classname);
	
	if (p != null)
	    return p;
	
	BufferedInputStream file;
	String cname = classname.replace('.', '/').concat(".class");
	
	// iterate through all components and look for .class file
	for (int i = 0; i < pathcomps.size(); i++)
	{
	    long fileSize;
	    String compname = (String)pathcomps.elementAt(i);

	    try
	    {
		// a zip file ?
		ZipFile  zip   = new ZipFile(compname);
		ZipEntry entry = zip.getEntry(cname);
		
		// avoid throwing and catching a null pointer exception
		if (entry == null)
		    throw new Exception("not in zip file");
		
		file = new BufferedInputStream(zip.getInputStream(entry));
		fileSize = entry.getSize();
		
		if (fileSize < 0)
		    throw new Exception("unsized zip entry for " +cname);
	    }
	    catch (Exception ie0)
	    {
		try
		{
		    String fn  = compname.concat("/").concat(cname);
		    File f = new File(fn);
		    
		    // a regular file
		    file = new BufferedInputStream(new FileInputStream(f));
		    fileSize = f.length();
		    
		}
		catch (IOException ie1)
		{
		    continue;
		} 
	    }
	    
	    // now try to read the file
	    if (fileSize > (1 * 1024 * 1024))
		throw new Exception("Class file is too large for " +cname);
	    
	    p = new byte[(int)fileSize];
	    int off = 0;
	    while (off < p.length)
	    {
		int rc;
		rc = file.read(p, off, p.length - off);
		if (rc <= 0)
		    throw new InternalError("Read returned " +rc+ " that is bogus.");
		off += rc;
	    }
	    
	    classcache.put(cname, p);
	    return p;
	}
	
	throw new Exception("Couldn't resolve class `"  // XXX better exception
			    + classname + "'\n" 
			    + " in CLASSPATH " + this);
    }
    
    /**
     * convert to printable form
     * 
     * @return String describing this class path
     */
    public String toString()
    {
	return pathcomps.toString();
    }
}
