/* 
 * LineListTokenizer.java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package ants.util;

import java.io.StreamTokenizer;
import java.io.FileNotFoundException;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.Reader;

import ants.core.Assert;

/**
 * File reading utility. Returns each useful line of text as an array
 * of strings, one string per "word."  Comments are ignored and quoted
 * strings are handled (the quotes are stripped off and the quoted
 * string is returned as a single word).
 * <p>
 * Max of 256 words per line (see MAXLIST).
 *
 * @author David Wetherall
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class LineListTokenizer
    extends StreamTokenizer 
{
    /**
     * Sorry, no more than 256 elements per line.
     */
    private final static int MAXLIST = 256;
    
    private int ll_nextToken()
	throws LineListException
    {
	try
	{
	    return this.nextToken();
	}
	catch (java.io.IOException ioe)
	{
	    throw new LineListException("ioe: " +ioe.getMessage());
	}
    }
    
    /**
     * Parse the next non-empty line into an array of strings.
     * Result is null when EOF, and otherwise has at least one element.
     * <p>
     * XXX rename to <code>nextLine()</code>
     *
     * @return an array of tokens from the line.
     */
    public String[] list() 
	throws LineListException
    {
	final String[] list;
	boolean eof;
	boolean eol;
	int length;
	
	list = new String[MAXLIST];
	eof = false;    
	eol = false;
	length = 0;
	
	// Loop until end-of-file, or end of a non-blank line.
	while ((!eof)
	       && ((!eol)
		   || (length == 0)))
	{
	    // Reset eol incase we just read a blank line.
	    eol = false;
	    
	    int t = ll_nextToken();
	    
	    if (t == TT_EOF)
		eof = true;
	    else if (t == TT_EOL)
		eol = true;
	    // numbers not expected but handle them anyhow
	    else if (t == TT_NUMBER)
		list[length++] = Double.toString(nval);
	    else
	    {
		if( sval != null )
		    list[length++] = sval;
	    }
	    
	    if (length == MAXLIST)
		throw new LineListException("Line too long");
	}
	
	if (length == 0)
	{
	    Assert.assert(eof);
	    return null;
	}
	
	String[] result = new String[length];
	System.arraycopy(list, 0, result, 0, length);
	return result;
    }
    
    
    /**
     * Create a new LineListTokenizer for the given file.
     *
     * @param file the file to use as input.
     * @exception FileNotFoundException if the given file cannot be opened for reading.
     */
    public LineListTokenizer(String file)
	throws FileNotFoundException
    {
	this(new InputStreamReader(new FileInputStream(file)));
    }
    
    
    /**
     * Create a new LineListTokenizer from the given readable
     * Stream.
     *
     * @param input the Reader to stream input from.
     */
    public LineListTokenizer(Reader input)
    {
	super(input);
	
	resetSyntax();
	whitespaceChars(0, ' ');
	wordChars('0', '9');
	wordChars('-', '-');
	wordChars('.', '.');
	wordChars(',', ',');
	wordChars(':', ':');
	wordChars('=', '=');
	wordChars('/', '/');
	wordChars('A', 'Z');
	wordChars('a', 'z');
	wordChars('_', '_');
	wordChars('*','*');
	commentChar('#');
	quoteChar('\'');
	quoteChar('\"');    
	eolIsSignificant(true);
    }
}
