/* 
 * MobileApplication.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */

/*
 * The Mobile test application for ANTS
 *
 * This is a simple application that uses a mobile protocol implementation to
 * forward packets from a home router to a foreign router, and finally to the
 * host.  The application takes arguments specifying the layout of the network
 * and which instance of the application should act as the server.  When the
 * mobile is on the home network, the router will just deliver packets directly
 * to it.  When the host is on the foreign network it will send out update
 * packets to store forwarding information on the foreign and home routers.
 * Any mobile capsules sent from a server will then use the forwarding info
 * on the routers to direct itself to the new location of the mobile.
 *
 *
 * Network Layout:
 *
 *                                       server
 *                                         |
 * mobile@foreign -- foreign router -- (internet) -- home router -- mobile@home
 *
 *
 * Output:
 *
 *  The output at the moment is pretty basic, it just prints out a message
 *  whenever it receives a capsule.  The thing to look for is that the 'from'
 *  address changes from the mobile@home address to the mobile@foreign address,
 *  and then changes back again, after the '*** Switching to' messages.
 */

package apps.mobile;

import edu.utah.janos.nodeos.AddressFormatException;

import ants.core.Application;
import ants.core.ByteArray;
import ants.core.Capsule;
import ants.core.Node;
import ants.core.NodeAddress;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;


/**
 * primitive Mobile test application
 *
 * @author David Wetherall
 */
public class MobileApplication
    extends Application
    implements Runnable
{
    int received;
    int iter;
    int mobile;
    int home;
    int foreign;
    int alternate;
    int udelay;
    int switchDelay;
    int switchCount;
    volatile boolean away;
    
    private final int LOG_LEVEL = 1;
    
    final public static String[] defaults = { "-update", "1000",
					      "-iter", "10000",
					      "-switchcount", "10000",
					      "-switch", "8" };
    
    /*
     * Class that will send any update capsules when we're away
     */
    class MAUpdate
	implements Runnable
    {
	MobileApplication mapp;
	
	private final int LOG_LEVEL = 1;
	
	/**
	 * Loop forever and send update capsules whenever we're away
	 */
	public void run()
	{
	    int switcheroo = 1;
	    int loop = 0;
	    
	    /* Loop for a while */
	    while( iter >= 0 )
	    {
		try
		{
		    /* Sleep for our update interval time */
		    mapp.thisNode().sleep(mapp.udelay);
		}
		catch (InterruptedException e)
		{
		    continue;
		}
		
				/* Switch from mobile to host every now and then... */
		if( (switchCount > 0) &&
		    (switcheroo % switchDelay == 0) )
		{
		    switchCount--;
		    mapp.away = !mapp.away;
		    mapp.thisNode().log(LOG_LEVEL,
					"*** Switching to " +
					(mapp.away ?
					 "away" :
					 "home"));
		}
		switcheroo++;
		if (!mapp.away) continue;
		
				/*
				 * The mobile host is on the foreign network, send
				 * out an update packet
				 */
		mapp.send(new MobileUpdateCapsule(mapp.home,
						  mapp.foreign,
						  mapp.alternate));
				/* Track the number of times we've sent an update */
		loop++;
		mapp.thisNode().log(LOG_LEVEL,
				    "*** " + loop + " updates");
	    }
	}
	
	/**
	 * Create an MAUpdate 
	 */
	MAUpdate(MobileApplication mapp)
	{
	    this.mapp = mapp;
	}
    }
    
    synchronized public void receive(Capsule cap) {
	super.receive(cap);
	
	if (cap instanceof MobileCapsule)
	{
	    /*
	     * We received a packet!  Dump out some information
	     * about it.
	     */
	    received++;
	    iter--;
	    if( iter <= 0 )
		thisNode().shutdown();
	    MobileCapsule mcap = (MobileCapsule)cap;
	    thisNode().log(LOG_LEVEL,
			   "Capsule #" + received
			   + " ("
			   + mcap.data.getByte(0)
			   + ","
			   + (char)mcap.data.getByte(1)
			   + ") from "
			   /*
			    * The next field will be set to the
			    * destination address, which will
			    * either be the mobile host, or the
			    * foreign network address for the
			    * mobile.
			    */
			   + NodeAddress.toString(mcap.next)
			   );
	}
	else
	{
	    thisNode().log(LOG_LEVEL,
			   "Received non mobile capsule "
			   + cap);
	}
    }
    
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	try
	{
	    for (int i = 0; i < k.length(); i++) {
		if (k.key(i).equals("-mobile")) {
		    mobile = NodeAddress.
			fromString(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-home")) {
		    home = NodeAddress.
			fromString(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-foreign")) {
		    foreign = NodeAddress.
			fromString(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-alternate")) {
		    alternate = NodeAddress.
			fromString(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-update")) {
		    udelay = Integer.parseInt(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-iter")) {
		    iter = Integer.parseInt(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-switch")) {
		    switchDelay =
			Integer.parseInt(k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-switchcount")) {
		    switchCount =
			Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
	    }
	}
	catch( AddressFormatException afe )
	{
	    throw new InvalidKeyArgsException("HACK HACK HACK: "
					      + afe.getMessage());
	}
	
	super.setArgs(k);
    }
    
    /**
     * Send packets to the mobile host, hopefully they get routed correctly
     */
    public void run()
    {
	ByteArray ba;
	byte count = 0;
	
	/* Construct some data to send in a packet */
	ba = new ByteArray(2);
	ba.setByte(1, (byte)'a');
	while( iter != 0 )
	{
	    iter--;
	    /* Add something that will change for each packet */
	    ba.setByte(0, count);
	    try
	    {
				/* Sleep for a bit and then send a packet */
		thisNode().sleep( 50 );
		this.send( new MobileCapsule(mobile,
					     (short)getPort(),
					     (short)getPort(),
					     ba) );
	    }
	    catch( InterruptedException ie ) 
	    {
		;
	    }
	    count++;
	}
    }
    
    public void start() 
	throws Exception 
    {
	thisNode().register(new MobileProtocol());
	
	if (mobile == thisNode().getAddress())
	{
	    /*
	     * We're the mobile host, startup the update thread
	     * that will act as the client and change from home to
	     * away, and send out MobileUpdateCapsules
	     */
	    thisNode().threadStart(new MAUpdate(this));
	} else {
	    /*
	     * We're the source host, just start a thread for this
	     * object that constantly sends objects off to the
	     * mobile host.
	     */
	    thisNode().threadStart(this);
	}
    }
}
