/* 
 * MobileUpdateCapsule.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package apps.mobile;

import ants.core.*;

/**
 * Mobile update capsule processor - This will update forwarding information
 * on routers so that MobileCapsules can direct themselves to the proper
 * destination
 *
 * @author David Wetherall
 */

public class MobileUpdateCapsule
    extends Capsule
{
    final private static byte[] MID = findMID("apps.mobile"
					      + ".MobileUpdateCapsule");
    protected byte[] mid()
    {
	return MID;
    }
    
    final private static byte[] PID = findPID("apps.mobile"
					      + ".MobileUpdateCapsule");
    protected byte[] pid()
    {
	return PID;
    }
    
    /**
     * The next place to ship a packet
     */
    int next;
    
    /**
     * The forwarding address to leave behind on the router
     */
    int forward;
    
    public int length()
    {
	return super.length() + Xdr.INT + Xdr.INT;
    }
    
    public Xdr encode()
    {
	Xdr xdr = super.encode();
	
	xdr.PUT(this.next);
	xdr.PUT(this.forward);
	
	return xdr;
    }
    
    public Xdr decode()
    {
	Xdr xdr = super.decode();
	
	this.next = xdr.INT();
	this.forward = xdr.INT();
	
	return xdr;
    }
    
    /**
     * Update the forwarding information on the foreign and home routers.
     * The MobileCapsules then use this forwarding information to get to
     * the right location.
     */
    public boolean evaluate(Node n)
    {
	boolean retval = true;
	
	if( n.getAddress() == this.next )
	{
	    /*
	     * Since the capsule is sent from the mobile on the
	     * foreign network we will hit the foreign router
	     * first.  So we update its forwarding info to send
	     * stuff for the mobile's original address, to the
	     * one used on the foreign network.  Next, the packet
	     * will be sent on to the home router, which will
	     * update the forwarding information to send packets
	     * to the foreign router.
	     */
	    
	    /*
	     * Update the Node cache with our forwarding
	     * information.  We use the mobile's address as the key
	     * and then store the forwarding address.
	     * MobileCapsules can then lookup forwarding info based
	     * on their destination address.
	     */
	    n.getCache().put(this.getSrc(),
			     new Integer(this.forward),
			     5);
	    if( n.getAddress() != this.getDst() )
	    {
				/*
				 * We're on the foreign router, switch the
				 * packet data around so that the home node
				 * will forward to this router.  Finally, send
				 * the packet on to the home router.
				 */
		this.forward = this.next;
		this.next = this.getDst();
		retval = n.routeForNode(this, this.next);
	    }
	}
	else
	{
	    /* Pass along packets until they hit the next addr */
	    retval = n.routeForNode(this, this.next);
	}
	return retval;
    }
    
    public MobileUpdateCapsule()
    {
    }
    
    /**
     * Construct an update capsule
     *
     * @param dst The address of the home router
     * @param next The address of the foreign router
     * @param forward The address of the mobile on the foreign network
     */
    public MobileUpdateCapsule(int dst, int next, int forward)
    {
	this.setDst(dst);
	this.next = next;
	this.forward = forward;
    }
    
}
