/* 
 * MulticastApplication.Java
 *
 * Copyright (c) 1999, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *  
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package apps.multicast;

import ants.core.*;
import ants.util.*;

/**
 * Thread to send out periodic subscribe requests for the multicast group.
 */
class MASubscribe
    implements Runnable 
{
    MulticastApplication mapp;
    
    public void run() {
	while (true) {
	    try {
		Thread.sleep(mapp.sdelay);
	    } catch (InterruptedException e) {
		continue;
	    }
	    
	    int target = mapp.target;
	    mapp.thisNode().log(0, "Sending subscribe capsule.");
	    mapp.send(new MulticastSubscribeCapsule(mapp.target, mapp.group,
						    (short)mapp.getPort()));
	}
    }
    
    MASubscribe(MulticastApplication mapp) {
	this.mapp = mapp;
    }
}

/**
 * Thread to send out periodic data capsules to the multicast group.
 */
class MASend
    implements Runnable
{
    MulticastApplication mapp;
    int count = 0;
    
    public void run() {
	while (true) {
	    try {
		Thread.sleep(mapp.sdelay);
	    } catch (InterruptedException e) {
		continue;
	    }
	    
	    /* Increment multicast count. */
	    count++;
	    
	    int target = mapp.target;
	    mapp.thisNode().log(0, "Sending multicast capsule # "
			       + count);
	    byte[] data = { 0x01, 0x02, 0x03, 0x04,
			    (byte) (count >> 24),
			    (byte) (count >> 16),
			    (byte) (count >> 8),
			    (byte) (count) };
	    mapp.send(new MulticastCapsule(mapp.group, mapp.target,
					   (short)mapp.getPort(),
					   new ByteArray(data)));
	}
    }
    
    MASend(MulticastApplication mapp) {
	this.mapp = mapp;
    }
}

/**
 * The Multicast test application for ANTSR(R).
 *
 * Input:
 *   -delay n	Number of milliseconds to wait between multicast subscriptions
 *		or number of milliseconds to wait between multicasts.
 *   -target a	Address of multicast sender (target for subscriptions).
 *   -group a	Multicast group "address".
 *
 * Output:
 *   This program generates output indicating when capsules are received.
 *
 * @author David Wetherall
 */
public class MulticastApplication
    extends Application
{
    int target;
    int group;
    int sdelay;
    int scount;
    
    final private static String[] DEFAULTS = { "-delay", "1000" };
    
    synchronized public void receive(Capsule cap) {
	super.receive(cap);
	
	if (cap instanceof MulticastCapsule) {
	    MulticastCapsule mc = (MulticastCapsule)cap;
	    
	    if (mc.data.length() < 8)
		thisNode().log(0, "Received bad multicast capsule!");
	    else {
		int count = ((int)mc.data.getByte(4)) << 24
		    | ((int)mc.data.getByte(5)) << 16
		    | ((int)mc.data.getByte(6)) << 8
		    | (mc.data.getByte(7) & 0xff);
		thisNode().log(0, "Received multicast capsule #" + count);
	    }
	} else if (cap instanceof MulticastSubscribeCapsule) {
	    scount++;
	    thisNode().log(0, "Received multicast subscription #" + scount);
	}
    }
    
    
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(DEFAULTS);
	
	try {
	    for (int i = 0; i < k.length(); i++) {
		if (k.key(i).equals("-target")) {
		    target = NodeAddress.fromString(
						    k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-group")) {
		    group = NodeAddress.fromString(
						   k.arg(i));
		    k.strike(i);
		} else if (k.key(i).equals("-delay")) {
		    sdelay = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
	    }
	}
	catch (Exception e) {
	    throw new InvalidKeyArgsException("HACK HACK HACK: "
					      + e.getMessage());
	}
	
	super.setArgs(k);
    }
    
    public void start() 
	throws Exception 
    {
	/* Register the multicast protocol. */
	thisNode().register(new MulticastProtocol());
	
	if (target == thisNode().getAddress()) {
	    thisNode().log(0, "Starting source thread...");
	    thisNode().threadStart(new MASend(this));
	} else {
	    thisNode().log(0, "Starting client thread...");
	    thisNode().threadStart(new MASubscribe(this));
	}
    }
    
    public MulticastApplication() {
	super();
    }
}

