/* 
 * MulticastSubscribeCapsule.Java
 *
 * Copyright (c) 1999, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 *  Copyright 1996 Massachusetts Institute of Technology
 * 
 *  Permission to use, copy, modify, distribute, and sell this software and
 *  its documentation for any purpose is hereby granted without fee, provided
 *  that the above copyright notice appear in all copies and that both that
 *  copyright notice and this permission notice appear in supporting
 *  documentation, and that the name of M.I.T. not be used in advertising or
 *  publicity pertaining to distribution of the software without specific,
 *  written prior permission.  M.I.T. makes no representations about the
 *  suitability of this software for any purpose.  It is provided "as is"
 *  without express or implied warranty.
 */
package apps.multicast;

import ants.core.*;

/**
 * Multicast subscription capsule processing
 *
 * @author David Wetherall
 */
public class MulticastSubscribeCapsule
    extends Capsule
{
    final private static byte[] MID	=
	findMID("apps.multicast"
		+ ".MulticastSubscribeCapsule");
    protected byte[] mid() { return MID; }
    
    final private static byte[] PID =
	findPID("apps.multicast"
		+ ".MulticastSubscribeCapsule");
    protected byte[] pid() { return PID; }
    
    final private static int SUPPRESS_INTERVAL = 1000; // milliseconds
    
    int group;	// The multicast group (address).
    int lastnode;	// Last node address.
    short destport;	// Destination port
    
    /**
     * Return the length of this capsule.
     */
    public int length() {
	return super.length() + Xdr.INT + Xdr.INT + Xdr.SHORT;
    }
    
    /**
     * Encode this capsule.
     */
    public Xdr encode() {
	Xdr xdr = super.encode();
	
	xdr.PUT(group);
	xdr.PUT(lastnode);
	xdr.PUT(destport);
	
	return xdr;
    }
    
    /**
     * Decode this capsule.
     */
    public Xdr decode() {
	Xdr xdr = super.decode();
	
	group = xdr.INT();
	lastnode = xdr.INT();
	destport = xdr.SHORT();
	
	return xdr;
    }
    
    /**
     * Process this capsule at the node.
     *
     * @param n The current node.
     */
    public boolean evaluate(Node n) {
	/* Create a key and try to get our soft state. */
	MulticastCapsuleCacheKey mkey
	    = new MulticastCapsuleCacheKey(group, getDst());
	MulticastCapsuleCacheValue mval
	    = (MulticastCapsuleCacheValue)n.getCache().get(mkey);
	n.log(0, "MulticastSubscribeCapsule: Trying to get state for group "
	      + NodeAddress.toString(group)
	      + ", dest " + NodeAddress.toString(getDst())
	      + ", key " + mkey.hashCode());
	
	if (mval == null) {
	    /* No soft state available -- either:
	     *   a) we haven't seen this multicast before, or
	     *   b) our soft state expired and was removed
	     */
	    
	    n.log(0, "MulticastSubscribeCapsule: No info at node "
		  + NodeAddress.toString(n.getAddress())
		  + "; (re)creating soft state...");
	    
	    /* Create new soft state. */
	    mval = new MulticastCapsuleCacheValue();
	} else {
	    /* We found our soft state! */
	    
	    n.log(0,
		  "MulticastSubscribeCapsule: Found info at node "
		  + NodeAddress.toString(n.getAddress()));
	}
	n.getCache().put(mkey, mval, 5);
	
	/* 
	 * If this capsule came from somewhere, record it in our soft
	 * state so that we can route the multicast to the subscriber.
	 */
	if (lastnode != 0) {
	    boolean found = false;
	    for (int i = 0;
		 mval.addrs != null && i < mval.addrs.length;
		 i++) {
		if (mval.addrs[i] == lastnode) {
		    found = true;
		    break;
		}
	    }
	    if (!found) {
		int len = (mval.addrs != null) ?
		    mval.addrs.length : 0;
		int[] nn = new int[len+1];
		if (len > 0) n.arraycopy(mval.addrs, 0, nn, 0, len);
		nn[len] = lastnode;
		mval.addrs = nn;
	    }
	}
	
	/*
	 * Record the current node as the last to handle this capsule.
	 */
	lastnode = n.getAddress();
	
	/*
	 * If the last update was less than a second ago, don't
	 * forward the multicast subscription request (to prevent an
	 * implosion at the sender).
	 */
	if (n.time() - mval.time < SUPPRESS_INTERVAL) return true;
	else mval.time = n.time();
	
	/* 
	 * Either deliver this capsule to the application, or forward
	 * it to the target (multicast sender).
	 */
	if (n.getAddress() == getDst())
	    return n.deliverToApp(this, destport);
	
	// n.log("onward to "
	//       + NodeAddress.toString(getDst()));
	
	return n.routeForNode(this, getDst());
    }
    
    public MulticastSubscribeCapsule()
    {
    }
    
    public MulticastSubscribeCapsule(int target, int group, short port)
    {
	setDst(target);
	this.group = group;
	this.lastnode = 0;
	this.destport = port;
    }
}

