/* 
 * PingApplication.Java
 *
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 * Based on the Ping example developed by David Wetherall for the
 * original ANTS.
 */
package apps.ping;

import ants.core.Application;
import ants.core.Capsule;
import ants.core.Protocol;
import ants.core.Node;
import ants.core.NodeAddress;
import ants.core.Xdr;
import ants.core.ByteArray;

import ants.util.KeyArgs;
import ants.util.InvalidKeyArgsException;

/**
 * This was once a version of djw's PingApplication.
 * <p>
 * Takes most parameters via command line arguments.
 * Accepts the following parameters (in addition to the
 * basic ones accepted by every Application).
 * <ul>
 *  <li> -target <addr>: Address for ping capsules to go to.
 *  <li> -iter <int>: Numer of ping capsules to send [10].
 *  <li> -interv <int>: Inter-capsule delay in ms [5].
 *  <li> -nbuf <int>: Number of packet buffers to provide to the flow [10]. !UNUSED
 *  <li> -replenish: Not used [false].
 *  <li> -prime <int>: Post prime packet delay.  Negative means no prime packet.  [-1].
 *  <li> -lossage <int>: Expected # packets lost [10]. 
 * </ul>
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class PingApplication
    extends Application
    implements Runnable
{
    private final static String[] defaults = {};
    
    /**
     * The logging level messages issued by this class
     * are issued at.  Gross hack to hide the grosser
     * hack that is `L'.
     */
    private final int LOG_LEVEL = L[6]; // XXX use Node logging interface
    
    /**
     * Not used.
     */
    private boolean replenish = false;
    
    private int prime = -1;
    
    private int target;
    private int rTotCount = 0;
    private int delay;
    private long beginTime;
    
    private int iter    = 200;
    private int lossage = 10;
    private int interv  = 5;
    private int nbuf    = 10;
    
    /**
     * Invoked when packets are received.
     */
    synchronized public void receive(Capsule unknownCap)
    {
	long now;
	super.receive(unknownCap);
	
	if (!(unknownCap instanceof PingCapsule))
	{
	    thisNode().log(LOG_LEVEL, "Non-ping capsule delivered?!!" +unknownCap);
	    thisNode().log(LOG_LEVEL, "PingCap.classLoader= " +PingCapsule.class.getClassLoader()+
			   "; other=" +unknownCap.getClass().getClassLoader());
	    return;
	}
	
	PingCapsule cap = (PingCapsule)unknownCap;
	
	now = thisNode().time();
	
	thisNode().log(LOG_LEVEL, "PingCapsule #" 
		       + cap.seqId
		       + " (" +rTotCount+ " received) delivered to App!");
	
	switch (rTotCount % 100)
	{
	case 0:
	    beginTime = now;
	    break;
	case 99:
	    double diff = 
		(((double)now - (double)beginTime) / 1000.0);
	    double thru = 100.0 / diff;
	    thisNode().log(LOG_LEVEL,
			   "Throughput is " + thru + " cap/sec");
	    
	    Xdr buf = new Xdr(cap.getData(), 0);
	    long lat = now - buf.LONG();
	    thisNode().log(LOG_LEVEL, "Latency is " + lat + "ms");
	    thisNode().log(LOG_LEVEL, "Received " + (rTotCount+1) + " packets");
	    break;
	}
	
	if (cap.seqId == -1) // don't count the prime capsule
	    return;
	
	rTotCount++;
	
	if (rTotCount >= iter - lossage)
	{
	    thisNode().log(LOG_LEVEL, "Received almost all capsules.  Exiting.");
	    thisNode().shutdown();
	}
    }
    
    public void run()
    {
	thisNode().log(LOG_LEVEL, "sending " 
		       +iter+ " capsules at " 
		       +interv+ "ms intervals");
	
	try
	{
	    // First, send a warmup capsule to get the protocols loaded at
	    // each node.
	    if (prime > 0)
	    {
		thisNode().log(LOG_LEVEL,
			       "Sending primer capsule (and waiting " +prime+ "ms).");
		sendCapsule(-1);
		thisNode().sleep(prime);
	    }
	    
	    for (int i=0; i < iter; i++)
	    {
		sendCapsule(i);
		thisNode().sleep(interv);
	    }
	}
	catch (InterruptedException _)
	{
	    // eh, whatever.
	    thisNode().log(LOG_LEVEL,
			   "PingApp Interrupted.  Try again.");
	}
    }
    
    private void sendCapsule(int i)
    {
	ByteArray buf = new ByteArray(Xdr.LONG);
	Xdr xdr = new Xdr(buf, 0);
	xdr.PUT(thisNode().time());
	PingCapsule c = new PingCapsule(getPort(), getPort(), target, buf);
	c.seqId = i;
	send(c);
    }
    
    /**
     * Parse command line arguments.
     * <p>
     * See the class description for a list of the arguments and
     * their meaning.
     */
    public void setArgs(KeyArgs k) 
	throws InvalidKeyArgsException
    {
	k.merge(defaults);
	
	try
	{
	    for (int i = 0; i < k.length(); i++)
	    {
		if (k.key(i).equals("-target"))
		{
		    target = NodeAddress.fromString(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-iter"))
		{
		    iter = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-interv"))
		{
		    interv = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-lossage"))
		{
		    lossage = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-nbuf"))
		{
		    nbuf = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-replenish"))
		{
		    replenish = true; // Boolean.parseBool(k.arg(i));
		    k.strike(i);
		}
		if (k.key(i).equals("-prime"))
		{
		    prime = Integer.parseInt(k.arg(i));
		    k.strike(i);
		}
	    }
	}
	catch (Exception e)
	{
	    throw new InvalidKeyArgsException("HACK HACK HACK: " +e.getMessage());
	}
	super.setArgs(k);
    }
    
    /**
     * Start the Non-GUI ping app.
     * <p>
     * Registers the PingProtocol with the local node, and then
     * starts a thread on <code>this.run()</code>.
     */
    public void start() 
	throws Exception 
    {
	thisNode().register(new PingProtocol());
	thisNode().threadStart(this);
    }
    
    /**
     * Construct a non-GUI ping app.
     * <p>
     * Doesn't do anything.
     */
    public PingApplication() 
	throws Exception 
    {
    }
}
