/*
 * StatsCapsule.Java
 *
 * Copyright (c) 1999, 2000, 2001 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 *         
 *
 */
package apps.stats;

import ants.core.Protocol;
import ants.core.Capsule;
import ants.core.Xdr;
import ants.core.Node;
import ants.core.ByteArray;
import ants.core.NodeAddress;
import ants.core.NodeCache;

import ants.core.RouteTable;
import edu.utah.janos.route.RouteEntry;

import java.util.Enumeration;
import java.util.Vector;

/**
 * Stats request capsule.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class StatsCapsule
    extends Capsule
{
    final private static byte[] MID = findMID("apps.stats" + ".StatsCapsule");
    final private static byte[] PID = findPID("apps.stats" + ".StatsCapsule");
    
    public int rootaddress;
    public long keyid;
    public short dpt;
    public byte depth;
    public long now;
    private final static int cachetime=180;
    
    public Vector branches;
    
    protected byte[] mid()
    {
	return MID;
    }
    
    protected byte[] pid()
    {
	return PID;
    }
    
    public int length()
    {
	return super.length()+Xdr.INT+Xdr.LONG+Xdr.SHORT+Xdr.BYTE+Xdr.LONG;
    }
    
    /**
     * Stats request capsules look like:<br>
     * 4	Address of the root of the tree<br>
     * 4	Unique key id of this request tree<br>
     * 2	Application port<br>
     * 1	Depth<br>
     * 4	Time capsule was sent
     */
    public Xdr encode()
    {
	Xdr xdr = super.encode();
	xdr.PUT(rootaddress);
	xdr.PUT(keyid);
	xdr.PUT(dpt);
	xdr.PUT((byte)(depth-1));
	xdr.PUT(now);
	return xdr;
    }
    
    public Xdr decode()
    {
	Xdr xdr = super.decode();
	rootaddress=xdr.INT();
	keyid=xdr.LONG();
	dpt=xdr.SHORT();
	depth=xdr.BYTE();
	now=xdr.LONG();
	return xdr;
    }
    
    /**
     * Does the actual process of branching the request packet
     * out to all neighboring nodes except those the have already
     * been visited.
     * <p>
     * Uses soft-state to persistantly keep track of which nodes
     * have already been visited and to keep from sending back to
     * root or the source of this request capsule
     * <p>
     * @return The number of requests branched out
     */
    private int branch(Node n)
    {
	int numbranches;
	Enumeration neighbors;
	int target;
	StatsCapsuleCacheKey skey;
	StatsCapsuleCacheValue sval;
	NodeCache cache;
	
	branches=new Vector();
	neighbors=n.getNeighborTable().getNeighbors();
	cache=n.getCache();
	// Add the source to the soft-state
	skey=new StatsCapsuleCacheKey(keyid,rootaddress,getPrevious());
	sval=new StatsCapsuleCacheValue();
	sval.time=n.time();
	cache.put(skey,sval,cachetime);
	// Add the root to the soft-state
	skey=new StatsCapsuleCacheKey(keyid,rootaddress,rootaddress);
	sval=new StatsCapsuleCacheValue();
	sval.time=n.time();
	cache.put(skey,sval,cachetime);
	numbranches=0;
	while (neighbors.hasMoreElements())
	{
	    target=((Integer)neighbors.nextElement()).intValue();
	    synchronized (cache)
	    {
		n.log(1,"*** Checking "+NodeAddress.toString(target));
		skey=new StatsCapsuleCacheKey(keyid,rootaddress,target);
		sval=(StatsCapsuleCacheValue)cache.get(skey);
		if (sval!=null)
		    continue;	// Already been visited
		else
		{
		    sval=new StatsCapsuleCacheValue();
		    sval.time=n.time();
		    cache.put(skey,sval,cachetime);
		}
	    }
	    n.log(1,"*** Sending stats request to "+
		  NodeAddress.toString(target));
	    StatsCapsule c=new StatsCapsule(target);
	    c.rootaddress=rootaddress;
	    c.keyid=keyid;
	    c.dpt=dpt;
	    c.depth=(byte)depth;
	    c.now=n.time();
	    c.prime(this);
	    n.routeForNode(c,target);
	    numbranches++;
	    branches.addElement(new Integer(target));
	}
	return numbranches;
    }
    
    /**
     * Encode information about this node to be returned to the root
     * of the stats
     * <p>
     * To change the statistics returned by the tree, change this function
     */
    private ByteArray encodeNodeInfo(Node n)
    {
	String addr = "<Unknown>";
	String proto = "<Unknown>";

	try
	{
	    addr = n.getVisibleAddressSpec();
	    proto = n.getVisibleProtocolSpec();
	}
	catch (Throwable securityException)
	{
	    // oh well, we're not allowed to see the phys addr on this node.
	}

	ByteArray buf = new ByteArray(Xdr.INT+
				      Xdr.STRING("Hostname")+Xdr.STRING(n.hostname())+
				      Xdr.STRING("Visible address spec")+Xdr.STRING(addr)+
				      Xdr.STRING("Visible protocol spec")+Xdr.STRING(proto));
	Xdr xdr = new Xdr(buf,0);
	xdr.PUT(3);		// Indicates three items of info
	xdr.PUT("Hostname");
	xdr.PUT(n.hostname());
	xdr.PUT("Visible address spec");
	xdr.PUT(addr);
	xdr.PUT("Visible protocol spec");
	xdr.PUT(proto);
	return buf;
    }
    
    public boolean evaluate(Node n)
    {
	int numbranches;
	if (n.getAddress()!=getSrc())
	{
	    StatsReturnCapsule retCap = new StatsReturnCapsule(dpt,dpt,getSrc(),encodeNodeInfo(n));
	    n.log(1,"*** StatsCapsule came from "+NodeAddress.toString(getPrevious()));
	    if (depth>0)
	    {
				// Branch out to neighboring nodes
		numbranches=branch(n);
	    } else {
		numbranches=0;
	    }
	    // Even if depth has expired, still need to notify
	    // root that I am here
	    n.log(1,"*** Sending return stats capsule to root "+
		  NodeAddress.toString(getSrc()));
	    // Return stats
	    retCap.infosource=n.getAddress();
	    retCap.parentnode=getPrevious();
	    retCap.latency=0;
	    retCap.capTime=n.time();
	    retCap.nodessentto=numbranches;
	    retCap.branches=branches;
	    retCap.prime(this);
	    if (!n.routeForNode(retCap,getSrc()))
		return false;
	    if (depth<=0)
	    {
		n.log(1,"I have been to the edge; and I'm not coming back (depth expired)");
	    }
	    return true;
	} else if (n.getAddress()!=getDst()) {
	    n.log(1,"*** Routing root stats request out to "+
		  NodeAddress.toString(getDst())+
		  " from root "+
		  NodeAddress.toString(rootaddress)+
		  " (key "+keyid+") at "+now);
	    return n.routeForNode(this,getDst());
	}
	return false;
    }
    
    public StatsCapsule()
    {
    }
    
    public StatsCapsule(int d)
    {
	setDst(d);
    }
}
