/*
 * StatsReturnCapsule.Java
 *
 * Copyright (c) 1999 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is part of the ANTS distribution.
 * 
 * Permission to use, copy, modify, distribute, and sell this
 * software and its documentation for any purpose is hereby granted
 * without fee, provided that (i) the above copyright notice(s)
 * appear in all copies and (ii) that both the copyright notice(s)
 * and this permission notice appear in supporting documentation,
 * and (iii) that the name(s) of the copyright holders not be used
 * in advertising or publicity pertaining to distribution of the
 * software without specific, written prior permission.  The
 * copyright holders make no representations about the suitability
 * of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 * 
 * The maintainers of this code ask, but do not require, that
 * improvements and bug fixes be emailed to ants-dist@cs.utah.edu
 */
package apps.stats;

import ants.core.Protocol;
import ants.core.DataCapsule;
import ants.core.Xdr;
import ants.core.Node;
import ants.core.ByteArray;
import ants.core.NodeAddress;

import java.util.Vector;

/**
 * Stats return capsule.
 *
 * @author <a href="mailto:janos-discuss@flux.cs.utah.edu">Utah Janos Team</a>
 */
public class StatsReturnCapsule
    extends DataCapsule
{
    final private static byte[] MID = findMID("apps.stats" + ".StatsReturnCapsule");
    final private static byte[] PID = findPID("apps.stats" + ".StatsReturnCapsule");

    public int infosource;
    public int parentnode;
    public long capTime;
    public long latency;
    public int nodessentto;
    public Vector branches=new Vector();

    protected byte[] mid()
    {
	return MID;
    }

    protected byte[] pid()
    {
	return PID;
    }

    public int length()
    {
	return super.length()+Xdr.INT+Xdr.INT+Xdr.LONG+Xdr.LONG+Xdr.INT+Xdr.INT*nodessentto;
    }

    /**
     * Return capsule looks like:<br>
     * 4	Address of the actual source of the information<br>
     * 4	Address of the node that requested stats from source node<br>
     * 4	Time capsule this is a response to was sent<br>
     * 4	Latency (0 if latency has yet to be figured)<br>
     * 4	Number of nodes the source of this return sent to
     * <p>
     * The data in the datacapsule contains strings which specify the
     * actual statistics of the node this is returning from
     */
    public Xdr encode()
    {
	int i;
	Xdr xdr = super.encode();
	xdr.PUT(infosource);
	xdr.PUT(parentnode);
	xdr.PUT(capTime);
	xdr.PUT(latency);
	xdr.PUT(nodessentto);
	for (i=0;i<nodessentto;i++)
	    xdr.PUT(((Integer)branches.elementAt(i)).intValue());
	return xdr;
    }

    public Xdr decode()
    {
	int i;
	Xdr xdr = super.decode();
	infosource=xdr.INT();
	parentnode=xdr.INT();
	capTime=xdr.LONG();
	latency=xdr.LONG();
	nodessentto=xdr.INT();
	for (i=0;i<nodessentto;i++)
	    branches.addElement(new Integer(xdr.INT()));
	return xdr;
    }

    public boolean evaluate(Node n)
    {
	// Check the latency
	if (latency==0)
	{
	    latency=(n.time()-capTime)/2;
	}
	if (n.getAddress()==getDst())
	{
	    return n.deliverToApp(this,dpt);
	} else {
	    return n.routeForNode(this,getDst());
	}
    }

    /**
     * Print out the information that is contained in this capsule
     */
    public void showInformation(Node n)
    {
	ByteArray buf=getData();
	Xdr xdr=new Xdr(buf,0);
	int i,num;
	n.log(0, "*** Recieved stats return capsule from "+
	      NodeAddress.toString(infosource)+
	      " with parent "+NodeAddress.toString(parentnode)+
	      " with latency of "+latency+" containing:");
	num=xdr.INT();
	for (i=0;i<num;i++)
	    n.log(0, "\t"+xdr.STRING()+": "+xdr.STRING());
    }

    public StatsReturnCapsule()
    {
    }

    public StatsReturnCapsule(int d)
    {
	setDst(d);
    }

    public StatsReturnCapsule(short sa, short da, int na, ByteArray d)
    {
	super(sa,da,na,d);
    }
}
